# RCAIDE/Library/Methods/Powertrain/Converters/Ducted_Fan.py
#  
# Created:  Jul 2023, M. Clarke
# Modified: Jan 2025, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import  RCAIDE
from RCAIDE.Framework.Core import Data ,redirect  
from RCAIDE.Framework.Analyses.Propulsion.Ducted_Fan_Design_Code import Ducted_Fan_Design_Code
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Rankine_Froude_Momentum_Theory.RFMT_performance                  import compute_ducted_fan_efficiency
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Blade_Element_Momentum_Theory.write_geometry                     import  write_geometry
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Blade_Element_Momentum_Theory.write_input_deck                   import  write_input_deck
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Blade_Element_Momentum_Theory.run_dfdc_analysis                  import  run_dfdc_analysis
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Blade_Element_Momentum_Theory.translate_conditions_to_dfdc_cases import  translate_conditions_to_dfdc_cases
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.Performance.Blade_Element_Momentum_Theory.read_results                       import  read_results 

# python imports   
from shutil import rmtree
from scipy import interpolate 
import os
import numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
#  design_ducted_fan
# ---------------------------------------------------------------------------------------------------------------------- 
def design_ducted_fan(ducted_fan, new_regression_results = False, keep_files = True): 
    """
    Designs and optimizes a ducted fan propulsor using either Rankine-Froude Momentum Theory or 
    Blade Element Momentum Theory (BEMT) with DFDC integration.

    Parameters
    ----------
    ducted_fan : Ducted_Fan
        Ducted fan component to be designed
    dfdc_bin_name : str, optional
        Name of DFDC executable, defaults to 'dfdc'
    new_regression_results : bool, optional
        Flag to generate new regression results, defaults to False
    keep_files : bool, optional
        Flag to keep DFDC input/output files, defaults to True

    Returns
    -------
    None
        Updates ducted_fan object with:
            - Blade geometry (twist, chord, radius distributions)
            - Performance characteristics at design point
            - Performance surrogate models (BEMT only)

    Notes
    -----
    Two fidelity levels are available:

    1. Rankine-Froude Momentum Theory:
        - Simple momentum theory calculations
        - Uses polynomial fits for performance coefficients
        - Quick preliminary design estimates

    2. Blade Element Momentum Theory (BEMT):
        - Detailed blade design using DFDC
        - Generates performance surrogate models
        - Accounts for 3D effects and losses
        - Requires external DFDC executable

    The BEMT method creates surrogate models for:
        - Thrust
        - Power
        - Efficiency
        - Torque
        - Thrust coefficient
        - Power coefficient

    **Major Assumptions**
        * Steady state operation
        * Incompressible flow for Rankine-Froude
        * No blade-wake interaction
        * Linear interpolation for surrogate models
        * US Standard Atmosphere 1976

    References
    ----------
    [1] Drela, M., "DFDC - Ducted Fan Design Code", MIT Aero & Astro, 2005
    [2] Esotec Developments, "DFDC Documentation", http://www.esotec.org/sw/DFDC.html

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.Ducted_Fan
    RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.compute_ducted_fan_performance
    """
    

    if ducted_fan.fidelity == 'Rankine_Froude_Momentum_Theory':
        
        omega     = ducted_fan.cruise.design_angular_velocity
        V         = ducted_fan.cruise.design_freestream_velocity 
        atmo      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        rho       = atmo.compute_values(ducted_fan.cruise.design_altitude,0.).density
        
        n, D, J, Cp, Ct, eta_p  = compute_ducted_fan_efficiency(ducted_fan, V, omega)

        T      = Ct * rho * (n**2)*(D**4) 
        P      = Cp * rho * (n**3)*(D**5)  
        Q      = P / omega
                
        ducted_fan.cruise.design_power              = P[0, 0]
        ducted_fan.cruise.design_efficiency         = eta_p 
        ducted_fan.cruise.design_thrust             = T[0, 0] 
        ducted_fan.cruise.design_torque             = Q[0, 0]
        ducted_fan.cruise.design_thrust_coefficient = Ct  
        ducted_fan.cruise.design_power_coefficient  = Cp 

    elif ducted_fan.fidelity == 'Blade_Element_Momentum_Theory': 
        if ducted_fan.cruise.design_altitude == None:
            raise AttributeError('design altitude not set')
        atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data       = atmosphere.compute_values(ducted_fan.cruise.design_altitude)      
        
        if ducted_fan.cruise.design_freestream_mach == None: 
            ducted_fan.cruise.design_freestream_mach = ducted_fan.cruise.design_freestream_velocity / atmo_data.speed_of_sound[0,0]  
    
        if ducted_fan.cruise.design_freestream_velocity == None: 
            ducted_fan.cruise.design_freestream_velocity = ducted_fan.cruise.design_freestream_mach * atmo_data.speed_of_sound[0,0]       
        
        if (ducted_fan.cruise.design_angular_velocity)  == None:
            raise AttributeError('design tip mach and/or angular velocity not set')
            
        dfdc_analysis                                   = Ducted_Fan_Design_Code() 
        dfdc_analysis.geometry                          = ducted_fan
        dfdc_analysis.settings.filenames.dfdc_bin_name  = ducted_fan.DFDC.bin_name
        dfdc_analysis.settings.new_regression_results   = new_regression_results
        dfdc_analysis.settings.keep_files               = keep_files
        
        # Get the root directory (one level above RCAIDE)
        current_dir = os.path.dirname(__file__)
        # Go up to the root directory containing RCAIDE
        root_dir = os.path.abspath(os.path.join(current_dir, *['..'] * 6))
    
        # Set the run folder path
        dfdc_analysis.settings.filenames.run_folder = os.path.join(root_dir, 'VnV', 'Verification', 'network_electric', 'dfdc_files')
        run_folder = dfdc_analysis.settings.filenames.run_folder
    
        # Ensure the directory exists
        os.makedirs(run_folder, exist_ok=True)
        run_script_path                                 = run_folder.rstrip('dfdc_files').rstrip('/')    
        deck_template                                   = dfdc_analysis.settings.filenames.deck_template 
        print_output                                    = dfdc_analysis.settings.print_output  
        dfdc_analysis.current_status.deck_file          = deck_template.format(1) 
     
        # translate conditions  
        translate_conditions_to_dfdc_cases(dfdc_analysis)  
        dfdc_analysis.settings.filenames.case  =  dfdc_analysis.geometry.tag +  '.case'
        # write the input files
        with redirect.folder(run_folder,force=False):
            write_geometry(dfdc_analysis,run_script_path)    
            write_input_deck(dfdc_analysis)   
    
            # RUN DFDC!
            _ = run_dfdc_analysis(dfdc_analysis,print_output)
    
        # translate results
        results = read_results(dfdc_analysis)
    
        if not dfdc_analysis.settings.keep_files:
            rmtree(run_folder) 
             
        # save blade geometry 
        ducted_fan.rotor.twist_distribution           = results.geometry.rotor_twist_distribution            
        ducted_fan.rotor.chord_distribution           = results.geometry.rotor_chord_distribution            
        ducted_fan.rotor.radius_distribution          = results.geometry.rotor_radius_distribution           
        ducted_fan.rotor.non_dim_radius_distribution  = results.geometry.rotor_non_dim_radius_distribution   
        ducted_fan.rotor.solidity_distribution        = results.geometry.rotor_solidity_distribution         
        ducted_fan.stator.twist_distribution          = results.geometry.stator_twist_distribution           
        ducted_fan.stator.chord_distribution          = results.geometry.stator_chord_distribution           
        ducted_fan.stator.radius_distribution         = results.geometry.stator_radius_distribution          
        ducted_fan.stator.non_dim_radius_distribution = results.geometry.stator_non_dim_radius_distribution  
        ducted_fan.stator.solidity_distribution       = results.geometry.stator_solidity_distribution   
    
        mach              = dfdc_analysis.training.mach       
        tip_mach          = dfdc_analysis.training.tip_mach         
        altitude          = dfdc_analysis.training.altitude/1000
    
        # create performance surrogates 
        raw_data           = results.performance
        thrust             = clean_data(raw_data.thrust,mach,tip_mach,altitude,raw_data.converged_solution)               
        power              = clean_data(raw_data.power,mach,tip_mach,altitude,raw_data.converged_solution)                
        efficiency         = clean_data(raw_data.efficiency,mach,tip_mach,altitude,raw_data.converged_solution)           
        torque             = clean_data(raw_data.torque,mach,tip_mach,altitude,raw_data.converged_solution)               
        thrust_coefficient = clean_data(raw_data.thrust_coefficient,mach,tip_mach,altitude,raw_data.converged_solution)   
        power_coefficient  = clean_data(raw_data.power_coefficient,mach,tip_mach,altitude,raw_data.converged_solution)     
        
        surrogates =  Data()
        surrogates.thrust              = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),thrust               ,method = 'linear',   bounds_error=False, fill_value=None)      
        surrogates.power               = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),power                ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.efficiency          = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),efficiency           ,method = 'linear',   bounds_error=False, fill_value=None)      
        surrogates.torque              = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),torque               ,method = 'linear',   bounds_error=False, fill_value=None)
        surrogates.thrust_coefficient  = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),thrust_coefficient   ,method = 'linear',   bounds_error=False, fill_value=None)      
        surrogates.power_coefficient   = interpolate.RegularGridInterpolator((mach,tip_mach,altitude),power_coefficient    ,method = 'linear',   bounds_error=False, fill_value=None) 
        
        ducted_fan.performance_surrogates =  surrogates 
        
        ducted_fan.cruise.design_thrust             = results.performance.design_thrust            
        ducted_fan.cruise.design_power              = results.performance.design_power             
        ducted_fan.cruise.design_efficiency         = results.performance.design_efficiency        
        ducted_fan.cruise.design_torque             = results.performance.design_power /  ducted_fan.cruise.design_angular_velocity   
        ducted_fan.cruise.design_thrust_coefficient = results.performance.design_thrust_coefficient
        ducted_fan.cruise.design_power_coefficient  = results.performance.design_power_coefficient  
        
    return

def clean_data(raw_data,mach,tip_mach,altitude,convergence_matrix):  
    cleaned_data =  np.zeros((len(mach),len(tip_mach),len(altitude))) 
    for i in range(len(mach)): 
        x = tip_mach
        y = altitude
        
        # mask invalid values
        array = np.ma.masked_invalid(raw_data[i])
        if np.all(array.mask ==False):
            cleaned_data[i, :, :] =  array.data
        else: 
            yy,xx   = np.meshgrid(y, x)
            x1      = xx[~array.mask]
            y1      = yy[~array.mask]
            newarr  = array[~array.mask] 
            points  = np.vstack((x1, y1)).T
            values1 = newarr.data
            cleaned_data[i, :, :]  = interpolate.griddata(points, values1, (xx, yy), method='cubic',  fill_value = -1E5)
             
    return cleaned_data




# RCAIDE/Library/Methods/Powertrain/Converters/Motor/append_motor_conditions.py
# 
# Created:  Jan 2025, M. Clarke, M. Guidotti

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_motor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_motor_conditions(motor,segment,energy_conditions): 

    """
    Initializes motor operating conditions for a mission segment.

    Parameters
    ----------
    motor : Converter
        Motor component (DC_Motor or PMSM_Motor) for which conditions are being initialized
    segment : Segment
        Mission segment containing the state conditions
    energy_conditions : Conditions
        Container for propulsor operating conditions

    Returns
    -------
    None
        Modifies energy_conditions in-place by adding motor-specific conditions

    Notes
    -----
    This function initializes arrays of zeros for key motor operating parameters during
    a mission segment. The conditions are stored in a nested structure under the motor's
    tag within energy_conditions.

    The following conditions are initialized:
        - torque: Motor output torque [N-m]
        - efficiency: Motor operating efficiency [-] 
        - current: Motor current draw [A]
        - voltage: Motor voltage [V]

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.DC_Motor
    RCAIDE.Library.Components.Powertrain.Converters.PMSM_Motor
    """


    ones_row    = segment.state.ones_row 
    energy_conditions.converters[motor.tag]                         = Conditions()
    energy_conditions.converters[motor.tag].inputs                  = Conditions()
    energy_conditions.converters[motor.tag].outputs                 = Conditions()
    energy_conditions.converters[motor.tag].efficiency              = 0. * ones_row(1)  
    energy_conditions.converters[motor.tag].inputs.voltage          = 0. * ones_row(1)
    energy_conditions.converters[motor.tag].inputs.current          = 0. * ones_row(1)
    energy_conditions.converters[motor.tag].outputs.work_done       = 0. * ones_row(1) 
    energy_conditions.converters[motor.tag].outputs.power           = 0. * ones_row(1)
    energy_conditions.converters[motor.tag].outputs.torque          = 0. * ones_row(1)  
    energy_conditions.converters[motor.tag].outputs.omega           = 0. * ones_row(1)
    return 




# RCAIDE/Methods/Performance/aircraft_aerodynamic_analysis.py
# 
# 
# Created:  Dec 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import  Data 
 
# Pacakge imports 
import numpy as np  

#------------------------------------------------------------------------------
# aircraft_aerodynamic_analysis
#------------------------------------------------------------------------------  
def aircraft_aerodynamic_analysis(aerodynamics_analysis = None,
                                  angle_of_attack_range = None,
                                  Mach_number_range= None,
                                  control_surface_deflection_range = np.array([[0]]),
                                  altitude = 0,
                                  delta_ISA=0):
    """
    Computes aerodynamic coefficients across ranges of angle of attack and Mach numbers using vortex lattice methods.
 
 
    Parameters
    --------
    vehicle : Vehicle
        The vehicle instance to be analyzed
    angle_of_attack_range : ndarray
        Array of angle of attack values to evaluate [radians]
    Mach_number_range : ndarray
        Array of Mach numbers to evaluate
    control_surface_deflection_range : ndarray, optional
        Array of control surface deflection angles [radians], default [[0]]
    altitude : float, optional
        Altitude for atmospheric properties [m], default 0
    delta_ISA : float, optional
        Temperature offset from ISA conditions [K], default 0
    use_surrogate : bool, optional
        Flag for using surrogate model in analysis, default True
    model_fuselage : bool, optional
        Flag for including fuselage effects, default True. Of note, fuselage modeling can 
        sometimes be difficult for VLM solvers.
 
    Returns
    --------
    results : Data
        Container of analysis results including:
            * Mach : ndarray
                Evaluated Mach numbers
            * alpha : ndarray
                Evaluated angles of attack [rad]
            * lift_coefficient : ndarray
                Computed lift coefficients
            * drag_coefficient : ndarray
                Computed drag coefficients
 
    Notes
    -----
    The function uses the US Standard Atmosphere 1976 model for atmospheric properties
    and evaluates aerodynamic coefficients using vortex lattice methods. Can use a surrogate model
    for faster evaluation or just direct evaluation of the aerodynamics. 
 
    **Major Assumptions**
        * Flow is steady and inviscid
        * Small angle approximations apply
        * Linear aerodynamics
        * Atmospheric properties follow US Standard Atmosphere 1976
 
    See Also
    --------
    RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method
    RCAIDE.Library.Attributes.Atmospheres.Earth.US_Standard_1976
    """

    #------------------------------------------------------------------------
    # setup flight conditions
    #------------------------------------------------------------------------   
    atmosphere     = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data      = atmosphere.compute_values(altitude,delta_ISA)
    P              = atmo_data.pressure 
    T              = atmo_data.temperature 
    rho            = atmo_data.density 
    a              = atmo_data.speed_of_sound 
    mu             = atmo_data.dynamic_viscosity
       
    # -----------------------------------------------------------------
    # Evaluate Without Surrogate
    # ----------------------------------------------------------------- 
    ctrl_pts = len(angle_of_attack_range[:, 0] )
    state                                         = RCAIDE.Framework.Mission.Common.State()
    state.conditions                              = RCAIDE.Framework.Mission.Common.Results() 
    state.conditions.freestream.density           = rho * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.dynamic_viscosity = mu  * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.temperature       = T   * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.pressure          = P   * np.ones_like(angle_of_attack_range)
    state.conditions.aerodynamics.angles.alpha    = angle_of_attack_range  
    state.conditions.aerodynamics.angles.beta     = angle_of_attack_range *0  
    state.conditions.freestream.u                 = angle_of_attack_range *0       
    state.conditions.freestream.v                 = angle_of_attack_range *0       
    state.conditions.freestream.w                 = angle_of_attack_range *0       
    state.conditions.static_stability.roll_rate   = angle_of_attack_range *0       
    state.conditions.static_stability.pitch_rate  = angle_of_attack_range *0 
    state.conditions.static_stability.yaw_rate    = angle_of_attack_range *0  
    state.conditions.expand_rows(ctrl_pts)
 
    CL_vals    = np.zeros((len(angle_of_attack_range),len(Mach_number_range)))  
    CD_vals    = np.zeros((len(angle_of_attack_range),len(Mach_number_range))) 
 
    state.analyses                                  =  Data()
    aerodynamics_analysis.initialize()            
    state.analyses.aerodynamics = aerodynamics_analysis 
    
    for i in range (len(Mach_number_range)):  
        state.conditions.freestream.mach_number                 = Mach_number_range[i, 0] * np.ones_like(angle_of_attack_range)
        state.conditions.freestream.velocity                    = Mach_number_range[i, 0] * a   * np.ones_like(angle_of_attack_range)   
        state.conditions.freestream.reynolds_number             = state.conditions.freestream.density * state.conditions.freestream.velocity / state.conditions.freestream.dynamic_viscosity 
        state.conditions.frames.inertial.velocity_vector[:,0]   = Mach_number_range[i, 0] * a[0, 0]   *  angle_of_attack_range[:, 0] 
        
     
        # ---------------------------------------------------------------------------------------
        # Evaluate With Surrogate
        # ---------------------------------------------------------------------------------------  
        _                 = state.analyses.aerodynamics.evaluate(state)        
        CL_vals[:,i]      = state.conditions.aerodynamics.coefficients.lift.total[:, 0]
        CD_vals[:,i]      = state.conditions.aerodynamics.coefficients.drag.total[:, 0] 

  
    results = Data(
        Mach              = Mach_number_range, 
        alpha             = angle_of_attack_range, 
        lift_coefficient  = CL_vals, 
        drag_coefficient  = CD_vals, 
    )  
          
    return results  



# RCAIDE/Library/Attributes/Coolants/__init__.py
# 

"""
RCAIDE Coolants module provides models and properties for various cooling fluids used in aircraft systems.

These models provide essential thermodynamic and physical properties needed for heat 
transfer calculations and thermal management system analysis, including Specific heat capacity,
Thermal conductivity, Density, Viscosity, and Heat transfer coefficients.

See Also
--------
RCAIDE.Library.Attributes.Cryogens : Cryogenic fluid properties
RCAIDE.Library.Components.Energy : Energy system components
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .Coolant       import Coolant
from .Glycol_Water  import Glycol_Water



# RCAIDE/Library/Methods/Powertrain/Converters/Engine/compute_throttle_from_power.py
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
 # RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core                                         import Units

# package imports
import numpy as np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  calculate_throttle_from_power
# ----------------------------------------------------------------------------------------------------------------------    
def compute_throttle_from_power(engine,conditions):
    """
    Computes engine throttle setting and fuel consumption based on required power output.

    Parameters
    ----------
    engine : RCAIDE.Library.Components.Propulsors
        Engine instance with the following attributes:
            - sea_level_power : float
                Maximum power output at sea level [W]
            - flat_rate_altitude : float
                Altitude below which power remains constant [m]
            - power_specific_fuel_consumption : float
                Power specific fuel consumption [kg/(W·s)]
    engine_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Engine operating conditions with:
            - power : numpy.ndarray
                Required power output [W]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - freestream.altitude : numpy.ndarray
                Current altitude [m]
            - freestream.delta_ISA : numpy.ndarray
                Temperature offset from standard atmosphere [K]

    Returns
    -------
    None

    Notes
    -----
    This function is the inverse of compute_power_from_throttle, calculating the
    required throttle setting to achieve a desired power output considering
    atmospheric conditions.

    **Major Assumptions**
        * Power varies linearly with density ratio above flat-rate altitude
        * Power remains constant below flat-rate altitude
        * Standard atmosphere conditions apply except for ISA temperature offset
        * Minimum power output is zero (negative values are clipped)

    **Theory**
    
    The power available is computed using:

    .. math::
        P_{available} = P_{SL} \\frac{\\sigma - 0.117}{0.883}

    Then throttle is determined by:

    .. math::
        \\text{throttle} = \\frac{P_{required}}{P_{available}}

    where:
        - :math:`P_{SL}` is sea-level power
        - :math:`\\sigma` is the density ratio
        - :math:`P_{required}` is the requested power output

    References
    ----------
    [1] Gudmundsson, S. (2014). General Aviation Aircraft Design: Applied Methods and Procedures. Butterworth-Heinemann.
    [2] Gagg and Ferrar

    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Engine.compute_power_from_throttle
    RCAIDE.Library.Attributes.Atmospheres.Earth.US_Standard_1976
    """

    # Unpack atmospheric conditions 
    delta_isa         = conditions.freestream.delta_ISA
    altitude          = conditions.freestream.altitude
    
    # Unpack engine operating conditions 
    engine_conditions = conditions.energy.converters[engine.tag] 
    PSLS              = engine.sea_level_power
    h_flat            = engine.flat_rate_altitude
    P                 = engine_conditions.power*1.0
    PSFC              = engine.power_specific_fuel_consumption
    
    altitude_virtual = altitude - h_flat        
    altitude_virtual[altitude_virtual<0.] = 0.   
    atmo             = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_values_0    = atmo.compute_values(0,0) 
    rho0             = atmo_values_0.density[0,0] 
    atmo_values      = atmo.compute_values(altitude_virtual,delta_isa) 
    rho              = atmo_values.density 

    #Compute density ratio 
    sigma        = rho / rho0 
    Pavailable   = PSLS * (sigma - 0.117) / 0.883        
    Pavailable[h_flat > altitude] = PSLS 
 
    # Compute throttle 
    throttle = P/Pavailable 
    P[P<0.] = 0. 

    # Compute fuel flow rate
    SFC             = PSFC* Units['lb/hp/hr']
    a               = np.zeros_like(altitude)
    fuel_flow_rate  = np.fmax(P*SFC,a)
    
    # Store outputs 
    engine_conditions.power_specific_fuel_consumption = PSFC
    engine_conditions.fuel_flow_rate                  = fuel_flow_rate
    engine_conditions.throttle                        = throttle

    return



# RCAIDE/Library/Plots/Performance/Mission/__init__.py
# 

"""
RCAIDE Mission Plotting Package

This package contains modules for visualizing mission-related data and analysis 
results in RCAIDE.

Notes
-----
The Mission plotting package provides visualization tools for:
    - Aircraft performance analysis
    - Flight trajectory visualization
    - Mission profile analysis
    - Flight condition studies

The modules focus on clear presentation of:
    - Time-varying parameters
    - Spatial trajectories
    - Flight condition data
    - Performance metrics

See Also
--------
RCAIDE.Library.Plots : Parent plotting package
RCAIDE.Library.Mission : Mission analysis tools
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .plot_aircraft_velocities     import plot_aircraft_velocities
from .plot_flight_conditions       import plot_flight_conditions
from .plot_flight_trajectory       import plot_flight_trajectory 


# RCAIDE/Library/Methods/Powertrain/Systems/compute_avionics_power_draw.py
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports
def compute_avionics_power_draw(avionics,bus,conditions):
    """
    Computes the power draw of an avionics system.
    
    Parameters
    ----------
    avionics : Avionics
        The avionics component with the following attributes:
            - power_draw : float
                Power consumption of the avionics component [W]
    avionics_conditions : Conditions
        Object to store avionics power conditions with the following attributes:
            - power : numpy.ndarray
                Array to store the computed power draw values [W]
    conditions : Conditions
        Object containing mission conditions (not directly used in this function)
    
    Returns
    -------
    None
        This function modifies the avionics_conditions.power array in-place.
    
    Notes
    -----
    This function assigns the constant power draw value from the avionics component
    to the power array in the avionics_conditions object. The power draw is assumed
    to be constant throughout the mission segment.
    
    For more complex avionics models, this function could be extended to calculate
    power draw based on operating mode, altitude, or other mission parameters.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Systems.append_avionics_conditions
    """
    bus_conditions                 = conditions.energy[bus.tag]
    avionics_conditions            = bus_conditions[avionics.tag]    
    avionics_conditions.power[:,0] = avionics.power_draw 
    bus_conditions.power_draw      += avionics_conditions.power*bus.power_split_ratio /bus.efficiency    
    return 


# RCAIDE/Library/Missions/Segments/Single_Point/Set_Speed_Set_Throttle.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """
    Initializes conditions for fixed speed and throttle analysis

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Flight altitude [m]
            - air_speed : float
                True airspeed [m/s]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - linear_acceleration_z : float
                Acceleration in z-direction [m/s^2]
            - roll_rate : float
                Aircraft roll rate [rad/s]
            - pitch_rate : float
                Aircraft pitch rate [rad/s]
            - yaw_rate : float
                Aircraft yaw rate [rad/s]
            - state:
                unknowns:
                    acceleration : array
                        X-direction acceleration [m/s^2]
                conditions : Data
                    State conditions container
                initials : Data, optional
                    Initial conditions from previous segment
    
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.position_vector [m]
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.acceleration_vector [m/s^2]
            - conditions.static_stability.roll_rate [rad/s]
            - conditions.static_stability.pitch_rate [rad/s]
            - conditions.static_stability.yaw_rate [rad/s]
    
    Notes
    -----
    This function sets up the initial conditions for a single point analysis with
    fixed speed and throttle setting. The x-acceleration is treated as an unknown
    to be solved for during the analysis.

    **Calculation Process**
        1. Check initial conditions
        2. Decompose velocity into components using sideslip angle:
                - v_x = V * cos(β)
                - v_y = V * sin(β)
            where:
                - V is true airspeed
                - β is sideslip angle
        3. Set position and altitude
        4. Initialize acceleration vector with unknown x-component
        5. Set angular rates

    **Major Assumptions**
        * Fixed throttle setting
        * Constant airspeed
        * Small angle approximations
        * Quasi-steady state
        * No lateral acceleration

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """      
    
    # unpack
    alt                     = segment.altitude
    air_speed               = segment.air_speed
    beta                    = segment.sideslip_angle  
    linear_acceleration_z   = segment.linear_acceleration_z
    roll_rate               = segment.roll_rate
    pitch_rate              = segment.pitch_rate
    yaw_rate                = segment.yaw_rate
    acceleration            = segment.state.unknowns.acceleration[0][0]
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    v_x  = np.cos(beta)*air_speed 
    v_y  = np.sin(beta)*air_speed
        
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x 
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y 
    segment.state.conditions.frames.inertial.acceleration_vector  = np.array([[acceleration,0.0,linear_acceleration_z]])  
    segment.state.conditions.static_stability.roll_rate[:,0]      = roll_rate         
    segment.state.conditions.static_stability.pitch_rate[:,0]     = pitch_rate
    segment.state.conditions.static_stability.yaw_rate[:,0]       = yaw_rate      

    
# ----------------------------------------------------------------------------------------------------------------------  
#  Unpack Unknowns 
# ----------------------------------------------------------------------------------------------------------------------  
def unpack_unknowns(segment):
    """ Unpacks the x accleration and body angle from the solver to the mission
    
        Assumptions:
        N/A
        
        Inputs:
            segment.state.unknowns:
                acceleration                        [meters/second^2]
                body_angle                          [radians]
            
        Outputs:
            segment.state.conditions:
                frames.inertial.acceleration_vector [meters/second^2]
                frames.body.inertial_rotations      [radians]

        Properties Used:
        N/A
                                
    """      
    
    # unpack unknowns  
    acceleration  = segment.state.unknowns.acceleration 
    segment.state.conditions.frames.inertial.acceleration_vector[0,0] = acceleration         


# RCAIDE/Library/Methods/Powertrain/Converters/Ducted_Fan/append_ducted_fan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_ducted_fan_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_ducted_fan_conditions(ducted_fan,segment,energy_conditions,noise_conditions=None): 

    """
    Initializes ducted fan operating conditions for a mission segment.
    
    Parameters
    ----------
    ducted_fan : Converter
        Ducted fan component for which conditions are being initialized
    segment : Segment
        Mission segment containing the state conditions
    energy_conditions : Conditions
        Container for energy system operating conditions
    
    Returns
    -------
    None
        Modifies energy_conditions in-place by adding ducted-fan-specific conditions
    
    Notes
    -----
    This function initializes arrays for key ducted fan operating parameters during
    a mission segment. The conditions are stored in a nested structure under the 
    ducted fan's tag within energy_conditions.
    
    The following conditions are initialized:
        - orientation : array(3)
            Fan orientation angles [rad]
        - commanded_thrust_vector_angle : array(1) 
            Commanded thrust vectoring angle [rad]
        - torque : array(1)
            Shaft torque [N-m]
        - throttle : array(1)
            Throttle setting [-]
        - thrust : array(1)
            Net thrust [N]
        - rpm : array(1)
            Rotor speed [RPM]
        - omega : array(1)
            Angular velocity [rad/s]
        - disc_loading : array(1)
            Thrust per unit disc area [N/m^2]
        - power_loading : array(1)
            Power per unit disc area [W/m^2]
        - tip_mach : array(1)
            Blade tip Mach number [-]
        - efficiency : array(1)
            Overall efficiency [-]
        - figure_of_merit : array(1)
            Hovering figure of merit [-]
        - power_coefficient : array(1)
            Non-dimensional power coefficient [-]
    
    **Major Assumptions**
        * All conditions except throttle are initialized as zero arrays
        * Throttle is initialized as ones array
        * Array length matches the segment state vector length
        * Conditions are stored using the ducted fan's tag as the dictionary key
    
    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.Ducted_Fan
    RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.compute_ducted_fan_performance
    """

    ones_row    = segment.state.ones_row 
    energy_conditions.converters[ducted_fan.tag]                               = Conditions()   
    energy_conditions.converters[ducted_fan.tag].orientation                   = 0. * ones_row(3) 
    energy_conditions.converters[ducted_fan.tag].commanded_thrust_vector_angle = 0. * ones_row(1) 
    energy_conditions.converters[ducted_fan.tag].torque                        = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].throttle                      = ones_row(1)
    energy_conditions.converters[ducted_fan.tag].thrust                        = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].rpm                           = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].omega                         = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].disc_loading                  = 0. * ones_row(1)                 
    energy_conditions.converters[ducted_fan.tag].power_loading                 = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].tip_mach                      = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].efficiency                    = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].figure_of_merit               = 0. * ones_row(1)
    energy_conditions.converters[ducted_fan.tag].power_coefficient             = 0. * ones_row(1)  
    return 





# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_propulsion_system_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
import  RCAIDE 
from RCAIDE.Framework.Core    import Units, Data
# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Propulsion System Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_propulsion_system_weight(vehicle,network):
    """
    Calculates the total propulsion system weight using Raymer's method, including subsystems.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - networks : list
                List of propulsion networks
            - fuselages : list
                List of fuselage components
            - flight_envelope : Data()
                Contains design_mach_number
            - mass_properties : Data()
                Contains max_zero_fuel
    network : RCAIDE.Network()
        Network component containing:
            - fuel_lines : list
                List of fuel line components with fuel tanks
            - propulsors : list
                List of propulsion components

    Returns
    -------
    output : Data()
        Propulsion system weight breakdown:
            - W_prop : float
                Total propulsion system weight [kg]
            - W_thrust_reverser : float
                Thrust reverser weight [kg]
            - W_starter : float
                Starter engine weight [kg]
            - W_engine_controls : float
                Engine controls weight [kg]
            - W_fuel_system : float
                Fuel system weight [kg]
            - W_nacelle : float
                Nacelle weight [kg]
            - W_engine : float
                Total dry engine weight [kg]
            - number_of_engines : int
                Number of engines
            - number_of_fuel_tanks : int
                Number of fuel tanks

    Notes
    -----
    This method calculates the complete propulsion system weight including engines,
    nacelles, fuel system, and all supporting systems using Raymer's correlations.

    **Major Assumptions**
        * Correlations based on conventional turbofan/turbojet installations
        * Engine controls scale with number of engines and fuselage length
        * Nacelle weight includes thrust reversers if applicable
        * Fuel system weight scales with fuel capacity and number of tanks
        * Starter weight scales with total engine weight

    **Theory**
    Key component weights are calculated using:
    .. math::
        W_{nacelle} = 0.6724K_{ng}L_n^{0.1}W_n^{0.294}N_{ult}^{0.119}W_{ec}^{0.611}N_{eng}^{0.984}S_n^{0.224}

    .. math::
        W_{fuel\_sys} = 1.07W_{fuel}^{0.58}N_{eng}^{0.43}M_{max}^{0.34}

    .. math::
        W_{engine} = 0.084BPR^{1.1}W_{eng}^{0.5}N_{eng}^{0.5}

    .. math::
        W_{engine\_controls} = 5N_{eng} + 0.8L_{eng}

    .. math::
        W_{starter} = 49.19\left(\frac{W_{eng}}{1000}\right)^{0.541}
    
    where:
        * :math:`K_{ng}` is a factor for the engine mount type
        * :math:`L_n` is the length of the nacelle
        * :math:`W_n` is the diameter of the nacelle
        * :math:`N_{ult}` is the ultimate load factor
        * :math:`W_{ec}` is the engine control weight
        * :math:`N_{eng}` is the number of engines
        * :math:`BPR` is the bypass ratio
        * :math:`W_{eng}` is the dry engine weight
        * :math:`L_{eng}` is the length of the engine
        * :math:`N_{eng}` is the number of engines
        * :math:`W_{fuel}` is the fuel weight
        * :math:`M_{max}` is the maximum Mach number
        * :math:`S_n` is the nacelle surface area

    References
    ----------
    [1] Raymer, D., "Aircraft Design: A Conceptual Approach", AIAA 
        Education Series, 2018. 

    See Also
    --------
    RCAIDE.Library.Methods.Weights.Correlation_Buildups.FLOPS.compute_jet_engine_weight
    RCAIDE.Library.Methods.Weights.Correlation_Buildups.FLOPS.compute_piston_engine_weight
    """

    NENG    =  0 
    WENG    =  0
    number_of_tanks =  0
    for network in  vehicle.networks:
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                number_of_tanks +=  1
            for propulsor in network.propulsors:
                if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet):
                    ref_propulsor = propulsor  
                    NENG  += 1
                    BPR =  propulsor.bypass_ratio
                    WENG   += 0.084 *  (propulsor.sealevel_static_thrust/Units.lbf)**1.1 * np.exp(-0.045*BPR) * Units.lbs # Raymer 3rd Edition eq. 10.4 
                if 'nacelle' in propulsor:
                    ref_nacelle =  propulsor.nacelle 
                    
    WFSYS           = compute_fuel_system_weight(vehicle, NENG)
    WNAC            = compute_nacelle_weight(vehicle,ref_nacelle, NENG, WENG)
    WEC, WSTART     = compute_misc_engine_weight(vehicle,NENG, WENG)
    WTHR            = 0
    WPRO            = WENG + WFSYS + WEC + WSTART + WTHR + WNAC

    output                      = Data()
    output.W_prop               = WPRO
    output.W_thrust_reverser    = WTHR
    output.W_starter            = WSTART
    output.W_engine_controls    = WEC
    output.W_fuel_system        = WFSYS
    output.W_nacelle            = WNAC
    output.W_engine             = WENG 
    output.number_of_engines    = NENG
    output.number_of_fuel_tanks = number_of_tanks  
    return output

def compute_nacelle_weight(vehicle,ref_nacelle, NENG, WENG):
    """ Calculates the nacelle weight based on the Raymer method
        Assumptions:
            1) All nacelles are identical
            2) The number of nacelles is the same as the number of engines 
            The engine weight is imported in Kg (The default RCAIDE mass unit)
        Source:
            Aircraft Design: A Conceptual Approach (2nd edition)

        Inputs:
            vehicle - data dictionary with vehicle properties                           [dimensionless]
                -.ultimate_load: ultimate load factor of aircraft
            nacelle  - data dictionary for the specific nacelle that is being estimated [dimensionless]
                -lenght: total length of engine                                         [m]
                -diameter: diameter of nacelle                                          [m]
            WENG    - dry engine weight                                                 [kg]


        Outputs:
            WNAC: nacelle weight                                                        [kg]

        Properties Used:
            N/A
    """ 
    Kng             = 1.017 # assuming the engine is pylon mounted
    Nlt             = ref_nacelle.length / Units.ft
    Nw              = ref_nacelle.diameter / Units.ft
    Wec             = 2.331 * (WENG/Units.lbs) ** 0.901 * 1.18
    Sn              = 2 * np.pi * Nw/2 * Nlt + np.pi * Nw**2/4 * 2
    WNAC            = 0.6724 * Kng * Nlt ** 0.1 * Nw ** 0.294 * vehicle.flight_envelope.ultimate_load ** 0.119 \
                      * Wec ** 0.611 * NENG ** 0.984 * Sn ** 0.224
    return WNAC * Units.lbs

def compute_misc_engine_weight(vehicle, NENG, WENG):
    """ Calculates the miscellaneous engine weight based on the Raymer method, electrical control system weight
        and starter engine weight
        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.fuselages['fuselage'].lengths.total: length of fuselage   [m]
            network    - data dictionary for the specific network that is being estimated [dimensionless]
                -.number_of_engines: number of engines

        Outputs:
            WEC: electrical engine control system weight                    [kg]
            WSTART: starter engine weight                                   [kg]

        Properties Used:
            N/A
    """

    L =  0 
    for fuselage in vehicle.fuselages:
        if L < fuselage.lengths.total: 
            total_length = fuselage.lengths.total / 2 # approximately the distance from cockpit to engine           
    Lec     = NENG * total_length / Units.ft
    WEC     = 5 * NENG + 0.8 * Lec
    WSTART  = 49.19*((WENG/Units.lbs)/1000)**0.541
    return WEC * Units.lbs, WSTART * Units.lbs
 
def compute_fuel_system_weight(vehicle, NENG):
    """ Calculates the weight of the fuel system based on the Raymer method
        Assumptions:

        Source:
            Aircraft Design: A Conceptual Approach

        Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.design_mach_number: design mach number
                -.mass_properties.max_zero_fuel: maximum zero fuel weight   [kg]

        Outputs:
            WFSYS: Fuel system weight                                       [kg]

        Properties Used:
            N/A
    """
    Nt = 0
    Vt = 0
    for network in vehicle.networks:
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                Nt +=1
                Vt += fuel_tank.volume / Units["gallon"]
    WFSYS = 2.405 * Vt**0.606 * 0.5 * Nt**0.5 
    return WFSYS * Units.lbs


# RCAIDE/Library/Plots/Energy/plot_battery_ragone_diagram.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_ragone_diagram(battery,
                          save_figure   = False, 
                          save_filename = "Ragone_Plot",
                          file_type     =  ".png",
                          width = 11, height = 7):
    """
    Creates a Ragone plot showing the relationship between specific power and specific energy of a battery.

    Parameters
    ----------
    battery : Battery
        RCAIDE battery object containing Ragone characteristics
        
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Ragone_Plot")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure handle containing the generated Ragone plot

    Notes
    -----
    The Ragone plot is a performance map showing the tradeoff between specific power 
    and specific energy in energy storage devices. The plot uses logarithmic scales 
    to display the wide range of values.
    
    **Major Assumptions**
    
    * Battery characteristics follow the Ragone relationship:
        P = const_1 * 10^(E * const_2)
    * Specific energy range is defined by upper and lower bounds
    * Values are converted to standard units (kW/kg and Wh/kg)

    **Definitions**
    
    'Specific Power'
        Power output per unit mass (kW/kg)
    'Specific Energy'
        Energy storage capacity per unit mass (Wh/kg)
    'Ragone Plot'
        Performance visualization showing the tradeoff between power and energy density
    """
 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
      
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
        
    axis_1 = plt.subplot(1,1,1)
    esp_plot = np.linspace(battery.cell.ragone.lower_bound, battery.cell.ragone.upper_bound,50)
    psp_plot = battery.cell.ragone.const_1*10**(esp_plot*battery.cell.ragone.const_2)
    axis_1.plot(esp_plot/(Units.Wh/Units.kg),psp_plot/(Units.kW/Units.kg), color = 'black', marker = ps.markers[0], linewidth = ps.line_width, label= battery.tag) 
    axis_1.set_ylabel('Specific Power (kW/kg)')
    axis_1.set_xlabel('Specific Energy (W-h/kg)')
    set_axes(axis_1)       
     
    # set title of plot 
    title_text    = 'Battery Ragone Plot'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return  fig 


# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/write_geometry.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Units 
from RCAIDE.Library.Methods.Geometry.Airfoil import  import_airfoil_geometry , compute_naca_4series
from .purge_files import purge_files   
import numpy as  np
 
# ---------------------------------------------------------------------------------------------------------------------- 
# Write Geometry 
# ---------------------------------------------------------------------------------------------------------------------- 
def write_geometry(dfdc_object, run_script_path):
    """
    Writes the ducted fan geometry to a case file for DFDC analysis.
    
    Parameters
    ----------
    dfdc_object : DFDCAnalysis
        Analysis object containing the following attributes:
            - geometry : DuctedFan
                Ducted fan geometry with the following attributes:
                    - tag : str
                        Identifier for the ducted fan
                    - length : float
                        Length of the ducted fan [m]
                    - hub_radius : float
                        Radius of the hub [m]
                    - tip_radius : float
                        Radius of the blade tip [m]
                    - blade_clearance : float
                        Clearance between blade tip and duct [m]
                    - number_of_rotor_blades : int
                        Number of blades on the rotor
                    - number_of_radial_stations : int
                        Number of radial stations for blade definition
                    - rotor : Data
                        Rotor properties
                            - percent_x_location : float
                                Axial location of rotor as fraction of length
                    - stator : Data
                        Stator properties
                            - percent_x_location : float
                                Axial location of stator as fraction of length
                    - cruise : Data
                        Cruise conditions
                            - design_angular_velocity : float
                                Design angular velocity [rad/s]
                            - design_freestream_velocity : float
                                Design freestream velocity [m/s]
                            - design_reference_velocity : float
                                Design reference velocity [m/s]
                    - hub_airfoil : Airfoil
                        Airfoil definition for hub
                    - duct_airfoil : Airfoil
                        Airfoil definition for duct
    run_script_path : str
        Path to the directory where the case file will be written
    
    Returns
    -------
    None
    
    Notes
    -----
    This function generates a DFDC case file with the following sections:
      1. Header information including operating conditions
      2. Hub geometry definition
      3. Separator between geometry components
      4. Duct geometry definition
    
    The function handles both NACA 4-series airfoils and custom airfoil
    definitions for the hub and duct. If no airfoil is specified, default
    NACA airfoils are used (0015 for hub, 2208 for duct).
    
    The blade geometry is defined with linearly varying chord and twist
    distributions from hub to tip.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.purge_files
    RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_geometry
    RCAIDE.Library.Methods.Geometry.Airfoil.compute_naca_4series
    """     
    # unpack inputs
    case_file   = dfdc_object.geometry.tag + '.case' 
    purge_files([case_file]) 
    geometry             = open(case_file,'w') 
    with open(case_file,'w') as geometry:
        make_header_text(dfdc_object,geometry)
        make_hub_text(dfdc_object,geometry)
        make_separator_text(dfdc_object,geometry)  
        make_duct_text(dfdc_object,geometry)
    return

def make_header_text(dfdc_object,geometry):  
    """This function writes the header using the template required for the DFDC executable to read
 
    """      
    header_1_text_template = \
'''DFDC Version  0.70 
Bladed rotor + actdisk stator test case                                             

OPER
!        Vinf         Vref          RPM          RPM
  {0}   {1}       {2}          0.0    
!         Rho          Vso          Rmu           Alt
   1.2260       340.00      0.17800E-04  0.00000E+00
!       XDwake        Nwake
  0.80000               20
!       Lwkrlx
            F
ENDOPER

AERO
!  #sections
     1
!   Xisection
  0.00000E+00
!       A0deg        dCLdA        CLmax         CLmin
  0.00000E+00   6.2800       1.5000      -1.0000    
!  dCLdAstall     dCLstall      Cmconst         Mcrit
  0.50000      0.20000      0.00000E+00  0.70000    
!       CDmin      CLCDmin     dCDdCL^2
  0.12000E-01  0.10000      0.50000E-02
!       REref        REexp
  0.20000E+06  0.35000    
ENDAERO

ROTOR
!       Xdisk        Nblds       NRsta
  {3}                {4}           {5}
!  #stations
    {6}
!           r        Chord         Beta
'''
    
    length                = dfdc_object.geometry.length
    x_disc_rotor          = dfdc_object.geometry.rotor.percent_x_location * length
    number_of_blades      = dfdc_object.geometry.number_of_rotor_blades 
    number_of_end_points  = dfdc_object.geometry.number_of_radial_stations + 1 
    number_of_stations    = dfdc_object.geometry.number_of_radial_stations  
    hub_radius            = dfdc_object.geometry.hub_radius
    clearance             = dfdc_object.geometry.blade_clearance
    design_RPM            = dfdc_object.geometry.cruise.design_angular_velocity /Units.rpm 
    tip_radius            = dfdc_object.geometry.tip_radius  
    V_inf                 = dfdc_object.geometry.cruise.design_freestream_velocity  
    V_ref                 = dfdc_object.geometry.cruise.design_reference_velocity  
    header_1_text         = header_1_text_template.format(V_inf,V_ref, design_RPM,x_disc_rotor,number_of_blades,number_of_end_points, number_of_stations)
   
    geometry.write(header_1_text)
    
    station_chords = np.linspace(0.4,0.3, number_of_stations) *tip_radius
    station_twists = np.linspace(77, 30, number_of_stations)
    station_radii  = np.linspace(hub_radius+clearance, tip_radius-clearance,number_of_stations )
    for i in range(number_of_stations): 
        station_radius = station_radii[i]
        station_chord  = station_chords[i] 
        station_twist  = station_twists[i]
        case_text = '  ' + format(station_radius, '.6f')+ "  " + format(station_chord, '.6f')+ "   " + format(station_twist, '.6f') + "\n" 
        geometry.write(case_text)
        
    header_2_text_template = \
'''ENDROTOR


AERO
!  #sections
     1
!   Xisection
  0.00000E+00
!       A0deg        dCLdA        CLmax         CLmin
  0.00000E+00   6.2800       1.0000      -1.5000    
!  dCLdAstall     dCLstall      Cmconst         Mcrit
  0.50000      0.20000      0.00000E+00  0.70000    
!       CDmin      CLCDmin     dCDdCL^2
  0.12000E-01 -0.10000      0.50000E-02
!       REref        REexp          TOC      dCDdCL^2
  0.20000E+06  0.35000      0.10000      0.20000E-01
ENDAERO

ACTDISK
! Xdisk   NRPdef
  {0}     15
! #stations
  3
! r     BGam
 0.02   -10.0
 0.06   -10.0
 0.10   -10.0
ENDACTDISK


DRAGOBJ
!  #pts
     4
!           x            r          CDA
  0.40000E-01  0.60000E-01  0.40000E-01
  0.40000E-01  0.80000E-01  0.35000E-01
  0.40000E-01  0.10000      0.30000E-01
  0.40000E-01  0.12000      0.30000E-01
ENDDRAGOBJ

'''
    x_disc_stator         = dfdc_object.geometry.stator.percent_x_location  * length  
    header_2_text         = header_2_text_template.format(x_disc_stator ) 
    geometry.write(header_2_text)     
    return  

def make_hub_text(dfdc_object,geometry):  
    """This function writes the rotor using the template required for the DFDC executable to read 
    """ 
    duct_header = \
'''GEOM
FatDuct + CB test case
''' 
    geometry.write(duct_header) 
    if len(dfdc_object.geometry.hub_airfoil) > 0:
        if type(dfdc_object.geometry.hub_airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
            NACA_Code =  dfdc_object.geometry.hub_airfoil.NACA_4_Series_code
            airfoil_geometry_data = compute_naca_4series(NACA_Code)    
        else: 
            airfoil_filename      = dfdc_object.geometry.hub_airfoil.coordinate_file
            airfoil_geometry_data = import_airfoil_geometry(airfoil_filename)
    else:  
        airfoil_geometry_data = compute_naca_4series('0015') 
    dim = len(airfoil_geometry_data.x_upper_surface)
    x_coords =  airfoil_geometry_data.x_upper_surface[::-1]
    y_coords =  airfoil_geometry_data.y_upper_surface[::-1]
     
    for i in range(dim):
        x_coord  = x_coords[i]* (dfdc_object.geometry.hub_radius / max(y_coords))
        y_coord  = y_coords[i]* (dfdc_object.geometry.hub_radius / max(y_coords))
        case_text = '     ' + format(x_coord, '.6f')+ "    " + format(y_coord, '.6f') + "\n" 
        geometry.write(case_text)
              
    return


def make_separator_text(dfdc_object,geometry):  
    """This function writes the operating conditions using the template required for the DFDC executable to read 
    """      
    separator_text = \
'''  999.0 999.0
'''
    geometry.write(separator_text) 
    return


def make_duct_text(dfdc_object,geometry):  
    """This function writes the operating conditions using the template required for the DFDC executable to read
 
    """      
    if len(dfdc_object.geometry.duct_airfoil) > 0: 
        if type(dfdc_object.geometry.duct_airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
            NACA_Code =  dfdc_object.geometry.duct_airfoil.NACA_4_Series_code
            airfoil_geometry_data = compute_naca_4series(NACA_Code)    
        else: 
            airfoil_filename      = dfdc_object.geometry.duct_airfoil.coordinate_file
            airfoil_geometry_data = import_airfoil_geometry(airfoil_filename)
    else:  
        airfoil_geometry_data= compute_naca_4series('2208')    
    dim = len(airfoil_geometry_data.x_coordinates)

    x_coords  = airfoil_geometry_data.x_coordinates*dfdc_object.geometry.length 
    y_coords  = airfoil_geometry_data.y_coordinates*dfdc_object.geometry.length     
    offset    = abs(min(y_coords)) + dfdc_object.geometry.tip_radius
    for i in range(dim):  
        case_text = '     ' + format(x_coords[i], '.6f')+ "    " + format(y_coords[i] + offset, '.6f') + "\n" 
        geometry.write(case_text) 
    end_text = \
'''ENDGEOM
'''               
    geometry.write(end_text) 
    return  
 



# RCAIDE/Library/Attributes/Planets/Planet.py
# 
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------  
#  Imports
# ----------------------------------------------------------------------------------------------------------------------  
 
from RCAIDE.Framework.Core import Data

# ----------------------------------------------------------------------------------------------------------------------  
#  Planet Class
# ----------------------------------------------------------------------------------------------------------------------   
class Planet(Data):
    """
    Base class for planetary bodies in RCAIDE. Provides fundamental planetary properties 
    and gravitational calculations.

    Attributes
    ----------
    mass : float
        Total mass of the planetary body in kg (defaults to 0.0)
    mean_radius : float
        Average radius of the planetary body in m (defaults to 0.0)

    Notes
    -----
    This class serves as the parent class for all specific planetary implementations.
    The default values are zero and should be overridden by child classes with 
    specific planetary properties.

    **Definitions**
    
    'Mean Radius'
        The average radius of a planetary body, accounting for polar flattening 
        and equatorial bulge
    
    'Mass'
        The total mass of the planetary body, including all components 
        (core, mantle, crust, atmosphere)

    See Also
    --------
    RCAIDE.Library.Attributes.Planets.Earth : Earth-specific implementation
    """
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """           
        self.mass              = 0.0  # kg
        self.mean_radius       = 0.0  # m 



# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/airframe_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import  RCAIDE
from RCAIDE.Framework.Core                 import Data  
from .clean_wing_noise                     import clean_wing_noise
from .landing_gear_noise                   import landing_gear_noise 
from .trailing_edge_flap_noise             import trailing_edge_flap_noise 
from RCAIDE.Library.Methods.Noise.Metrics  import A_weighting_metric  
from RCAIDE.Library.Methods.Noise.Common   import SPL_arithmetic 

# python imports 
import numpy as np

# ----------------------------------------------------------------------
#  Airframe Noise 
# ----------------------------------------------------------------------
def airframe_noise(microphone_locations,segment,config,settings):
    """ This computes the noise from different sources of the airframe for a given vehicle for a constant altitude flight. 

    Assumptions:
        Correlation based 
 
    Source:
       Fink, Martin R. "Noise component method for airframe noise." Journal of aircraft 16.10 (1979): 659-665.
               
    Inputs:
        vehicle	 - RCAIDE type vehicle

        includes these fields:
            S                          - Wing Area
            bw                         - Wing Span
            Sht                        - Horizontal tail area
            bht                        - Horizontal tail span
            Svt                        - Vertical tail area
            bvt                        - Vertical tail span
            deltaf                     - Flap deflection
            Sf                         - Flap area
            cf                         - Flap chord
            slots                      - Number of slots (Flap type)
            Dp                         - Main landing gear tyre diameter
            Hp                         - Main lading gear strut length
            Dn                         - Nose landing gear tyre diameter
            Hn                         - Nose landing gear strut length
            wheels                     - Number of wheels 
            
        noise segment - flight path data, containing:
            distance_vector             - distance from the source location to observer
            angle                       - polar angle from the source to the observer
            phi                         - azimuthal angle from the source to the observer


    Outputs: One Third Octave Band SPL [dB]
        SPL_wing                        - Sound Pressure Level of the clean wing
        SPLht                           - Sound Pressure Level of the horizontal tail
        SPLvt                           - Sound Pressure Level of the vertical tail
        SPL_flap                        - Sound Pressure Level of the flaps trailing edge
        SPL_slat                        - Sound Pressure Level of the slat leading edge
        SPL_main_landing_gear           - Sound Pressure Level og the main landing gear
        SPL_nose_landing_gear           - Sound Pressure Level of the nose landing gear

    Properties Used:
        N/A      
        
    """
    # Unpack conditions 
    velocity     = segment.conditions.freestream.velocity                  # aircraft velocity  
    noise_time   = segment.conditions.frames.inertial.time[:,0]            # time discretization

    # Generate array with the One Third Octave Band Center Frequencies
    frequency = settings.center_frequencies[5:]  
    num_f     = len(frequency) 
    n_cpts    = len(noise_time)  
    n_mic     = len(microphone_locations)
    
    # Unpack Geometry  
    slots      = 0 
    for wing in config.wings:
        if type(wing) == RCAIDE.Library.Components.Wings.Main_Wing:
            taper = wing.taper 
            Sw    = wing.areas.reference                
            bw    = wing.spans.projected               
            for cs in  wing.control_surfaces:  
                if type(cs) == RCAIDE.Library.Components.Wings.Control_Surfaces.Flap: 
                    deltaf                  = cs.deflection
                    flap_span               = (cs.span_fraction_end - cs.span_fraction_start) * bw
                    chord_root              = 2*Sw/bw/(1+taper) 
                    chord_tip               = taper * chord_root
                    delta_chord             = chord_tip - chord_root 
                    wing_chord_flap_start   = chord_root + delta_chord * cs.span_fraction_start 
                    wing_chord_flap_end     = chord_root + delta_chord * cs.span_fraction_end
                    flap_chord_start        = wing_chord_flap_start* cs.chord_fraction 
                    flap_chord_end          = wing_chord_flap_end* cs.chord_fraction   
                    cf                      = (flap_chord_start +flap_chord_end) /2  
                    Sf                      = flap_span * cf
                
                    # determining flap slot number
                    if cs.configuration_type   == 'single_slotted':
                        slots = 1
                    elif cs.configuration_type == 'double_slotted':
                        slots = 2
                    elif cs.configuration_type == 'triple_slotted':
                        slots = 3  
        elif type(wing) == RCAIDE.Library.Components.Wings.Horizontal_Tail: 
            Sht                     = wing.areas.reference    # horizontal tail area, sq.ft
            bht                     = wing.spans.projected    # horizontal tail span, ft
        elif type(wing) == RCAIDE.Library.Components.Wings.Vertical_Tail:  
            Svt                     = wing.areas.reference     # vertical tail area, sq.ft
            bvt                     = wing.spans.projected     # vertical tail span, ft
    
     
    Dp                 = 0
    Dn                 = 0
    main_wheels        = 0
    main_units         = 0
    Hp                 = 0
    Hn                 = 0 
    nose_wheels        = 0
    main_gear_extended = False
    nose_gear_extended = False
    
    for landing_gear in  config.landing_gears:
        if isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            Dp                 = landing_gear.tire_diameter           # MLG tyre diameter 
            Dn                 = landing_gear.strut_length            # NLG tyre diameter 
            main_wheels        = landing_gear.wheels                          # Number of wheels   
            main_gear_extended = landing_gear.gear_extended                   # Gear up or gear down 
            main_units         = landing_gear.units                           # Number of main units
        elif isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            Hp                 = landing_gear.tire_diameter                 # MLG strut length 
            Hn                 = landing_gear.strut_length                  # NLG strut length 
            nose_gear_extended = landing_gear.gear_extended                   # Gear up or gear down 
            nose_wheels        = landing_gear.wheels                          # Number of wheels   
      
    
    
    viscosity           = segment.conditions.freestream.kinematic_viscosity[:,0] 
    M                   = segment.conditions.freestream.mach_number 
    SPL_total_history   = np.zeros((n_cpts,n_mic,num_f)) 
    SPLt_dBA_history    = np.zeros((n_cpts,n_mic,num_f))   

    # Distance vector from the aircraft position in relation to the microphone coordinates [meters]
    distance          = np.linalg.norm(microphone_locations,axis = 1)
    
    altitude          = abs(microphone_locations[:,2])
    sideline_distance =  microphone_locations[:,1] 
    
    # Polar angle emission vector relatively to the aircraft to the microphone coordinates, [rad] 

    theta     =  np.zeros(n_mic)
    bool_1    = (microphone_locations[:,1] > 0) &  (microphone_locations[:,0] > 0)
    bool_2    = (microphone_locations[:,1] > 0) &  (microphone_locations[:,0] < 0)
    bool_3    = (microphone_locations[:,1] < 0) &  (microphone_locations[:,0] < 0)
    bool_4    = (microphone_locations[:,1] < 0) &  (microphone_locations[:,0] > 0)
    
    theta[bool_1] =  np.pi - np.arctan(microphone_locations[:,1]/microphone_locations[:,0])[bool_1]
    theta[bool_2] =  np.arctan(microphone_locations[:,1]/ abs(microphone_locations[:,0]))[bool_2]
    theta[bool_3] =  np.arctan(abs(microphone_locations[:,1])/ abs(microphone_locations[:,0]))[bool_3]
    theta[bool_4] =  np.pi - np.arctan(abs(microphone_locations[:,1])/ microphone_locations[:,0])[bool_4]
    
     # Azimuthal (sideline) angle emission vector relatively to the aircraft to the microphone coordinates, [rad] 
    phi   = np.arctan(sideline_distance/altitude)
    
    
    # START LOOP FOR EACH POSITION OF AIRCRAFT   
    for i in range(n_cpts):
        for j in range(n_mic):  
             
            SPL_wing = clean_wing_noise(Sw,bw,0,1, velocity[i,0],viscosity[i],M[i],phi[j],theta[j],distance[j],frequency)      # Wing Noise
            SPLht    = clean_wing_noise(Sht,bht,0,1, velocity[i,0],viscosity[i],M[i],phi[j],theta[j],distance[j],frequency)     # Horizontal Tail Noise
            SPLvt    = clean_wing_noise(Svt,bvt,0,0,velocity[i,0],viscosity[i],M[i],phi[j],theta[j],distance[j],frequency)     # Vertical Tail Noise
      
            # Flap noise 
            if deltaf==0:
                SPL_flap = np.zeros(num_f)
            else:
                SPL_flap = trailing_edge_flap_noise(Sf,cf,deltaf,slots,velocity[i,0],M[i],phi[j],theta[j],distance[j],frequency)  
    
            # Main landing gear noise     
            if main_gear_extended == False:  
                SPL_main_landing_gear = np.zeros(num_f)
            else:
                SPL_main_landing_gear = landing_gear_noise(Dp,Hp,main_wheels,M[i],velocity[i,0],phi[j],theta[j],distance[j],frequency)   
                if main_units>1: # Incoherent summation of each main landing gear unit
                    SPL_main_landing_gear = SPL_main_landing_gear+3*(main_units-1)      
                
            # Nose landing gear noise
            if nose_gear_extended == False:                  
                SPL_nose_landing_gear = np.zeros(num_f)
            else:
                SPL_nose_landing_gear = landing_gear_noise(Dn,Hn,nose_wheels,M[i],velocity[i,0],phi[j],theta[j],distance[j],frequency)   
               
            # Total Airframe Noise
            SPL_total = 10.*np.log10( 10.0**(0.1*SPL_wing)+ 10.0**(0.1*SPLht) + 10.0**(0.1*SPLvt) + 10.0**(0.1*SPL_flap) + 10.0**(0.1*SPL_main_landing_gear)+ 10.0**(0.1*SPL_nose_landing_gear))      
                
            SPL_total_history[i,j,:]             = SPL_total  
            
            # Calculation of dBA based on the sound pressure time history 
            SPLt_dBA_history[i,j,:] = A_weighting_metric(SPL_total,frequency) 
    
    # Pack Airframe Noise 
    airframe_noise                        = Data()  
    airframe_noise.SPL                    = SPL_arithmetic(SPL_total_history, sum_axis= 2)
    airframe_noise.SPL_1_3_spectrum       = SPL_total_history
    airframe_noise.SPL_dBA                = SPL_arithmetic(np.atleast_2d(SPLt_dBA_history), sum_axis= 2) 
    airframe_noise.noise_time             = noise_time 
    return airframe_noise



# RCAIDE/Library/Compoments/Powertrain/Sources/Fuel_Cell/Proton_Exchange_Membrane_Fuel_Cell.py
# 
# 
# Created:  Dec 2024, M. Guidotti and M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports
from RCAIDE.Framework.Core       import Data
from .Generic_Fuel_Cell_Stack    import  Generic_Fuel_Cell_Stack
from RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.compute_fuel_cell_performance import *
from RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.append_fuel_cell_conditions   import *

# ----------------------------------------------------------------------------------------------------------------------
#  Proton_Exchange_Membrane_Fuel_Cell
# ---------------------------------------------------------------------------------------------------------------------- 
class Proton_Exchange_Membrane_Fuel_Cell(Generic_Fuel_Cell_Stack):
    """
    Proton Exchange Membrane Fuel Cell class 
    """ 
    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None 

        Parameters:
        ----------
        type: string
            The type of PEM model to use, "LT" for low temperature, or "HT" for high temperature
        R: float 
            Universal gas constant (J / (mol*K))
        F: float 
            Faraday constant (C / mol)
        E_C: float 
            Activation energy of ORR (J)
        H2_molar_mass: float 
            Molar mass of H2 (kg/mol)
        O2_molar_mass: float 
            Molar mass of O2 (kg/mol)
        t_m: Length 
            The membrane thickness of the PEM cell 
        a_c: float 
            The catalyst specific area in cm^2 / mg Pt
        L_c: float 
            The catalyst platinum loading in mg Pt / cm^2
        A: Area 
            The membrane area 
        compressor_expander_module: compressor_expander_module 
            The compressor expander module of the air supply system 
        maximum_deg: float 
            The maximum voltage drop due to degradation of the fuel cell at EOL
        rated_current_density: float 
            The rated current density for pressure drop calculations (typically the current density at max power)
            Can be calculated using the evaluate_max_PD method
            Defaults to None 
        rated_power_density: float 
            The rated power density of the fuel cell 
            Defaults to None 
        rated_p_drop_fc: Pressure 
            The pressure drop at rated current density of the fuel cell cathode 
            Defaults to 0.240 bar 
        rated_p_drop_hum: Pressure 
            The pressure drop at rated current density of one side of the membrane humidifier
            Defaults to 0.025 bar
        gamma_para: float 
            The parasitic power draw divided by the gross cell power 
            Does not include compressor/ram-air heat exchanger coolant power
            Defaults to 0.03
        alpha: float 
            The charge transfer coefficient
            Used for both models, is between 0-1, but usually around 0.5 
            Defaults to 0.375 (matches reports and experimental data)
        gamma: float 
            The activation loss pressure dependency coefficient
            Generally 1 for HT-PEM 
            Defaults to 0.45 for LT-PEM
        lambda_eff: float 
            Fitting parameter used to calculate the conductivity of Nafion membranes 
            Only used if type is "LT"
            Values from 9-24 are reasonable 
            Defaults to 9.15 for fully humidifed air 
        c1: float 
            Membrane conductivity of HT PBI membrane at 100 C (ms/cm)
            Only used if type is "HT"
            defaults to 0.0435 for standard PBI membrane 
        c2: float 
            Membrane conductivity of HT PBI membrane at 200 C (ms/cm)
            Only used if type is "HT"
            defaults to 0.0636 for standard PBI membrane 
        i0ref: float 
            Reference current density for activation losses for HT PEM model (A / cm^2 Pt)
            Note that it is given per unit catalyst surface area, not per membrane area 
            defaults to 4e-8
        i0ref_P_ref: float 
            Pressure at which i0ref was determined  
            defaults to 1 bar 
        ioref_T_ref: float 
            Temperature at which i0ref was determined 
            defaults to 369.05 K
        i_lim_multiplier: float 
            The value which the limiting current is multiplied by for advanced PEM systems
        area_specific_mass: float 
            The mass per active membrane area of the fuel cell (kg/m2)
            
             
        References
        [1] Chilver-Stainer, James, et al. "Power output optimisation via arranging gas flow channels
            for low-temperature polymer electrolyte membrane fuel cell (PEMFC) for
            hydrogen-powered vehicles." Energies 16.9 (2023): 3722.
        
        """ 

        self.tag                                                               = 'pem_fuel_cell'        
        self.fuel_cell.Universal_gas_constant                                  = 8.31   # Universal gas constant (J / (mol*K))
        self.fuel_cell.Faraday_constant                                        = 96485  # Faraday constant (C / mol)
        self.fuel_cell.E_C                                                     = 66000  # Activation energy of ORR (J)
        self.fuel_cell.H2_molar_mass                                           = 2.0E-3 # Molar mass of H2 (kg/mol) 
        self.fuel_cell.O2_molar_mass                                           = 32E-3  # Molar mass of O2 (kg/mol) 
        self.fuel_cell.O2_mass_frac                                            = 0.233          
        self.fuel_cell.type                                                    = "LT"
        self.fuel_cell.specific_heat_capacity                                  = 903    # J/kg·K  # Chilver-Stainer, James, et al
        self.fuel_cell.t_m                                                     = 0.0024 #* Units.cm 
        self.fuel_cell.interface_area                                          = 50 #* Units.cm^2  # area of the fuel cell interface        
        self.fuel_cell.a_c                                                     = 98
        self.fuel_cell.L_c                                                     = 0.1 
        self.fuel_cell.compressor_expander_module                              = Data() 
        self.fuel_cell.compressor_expander_module.compressor_efficiency        = 0.71
        self.fuel_cell.compressor_expander_module.expander_efficiency          = 0.73 
        self.fuel_cell.compressor_expander_module.motor_efficiency             = 0.801025 
        self.fuel_cell.compressor_expander_module.generator_efficiency         = 1
        self.fuel_cell.compressor_expander_module.specific_weight              = None
        self.fuel_cell.compressor_expander_module.weight                       = 0 
        self.fuel_cell.maximum_deg                                             = 0
        self.fuel_cell.rated_current_density                                   = 0.1 # A/cm^2
        self.fuel_cell.rated_p_drop_fc                                         = 0.240 
        self.fuel_cell.rated_p_drop_hum                                        = 0.025
        self.fuel_cell.rated_H2_pressure                                       = 1   
        self.fuel_cell.rated_air_pressure                                      = 1  
        self.fuel_cell.gamma_para                                              = 0.03 
        self.fuel_cell.alpha                                                   = 0.375
        self.fuel_cell.gamma                                                   = 0.45
        self.fuel_cell.lambda_eff                                              = 24 
        self.fuel_cell.c1                                                      = 0.0435 
        self.fuel_cell.c2                                                      = 0.0636
        self.fuel_cell.stack_temperature                                       = 353.15 # Kelvin
        self.fuel_cell.air_excess_ratio                                        = 2
        self.fuel_cell.oxygen_relative_humidity                                = 1
        self.fuel_cell.i0ref                                                   = 9E-6
        self.fuel_cell.i0ref_P_ref                                             = 1 
        self.fuel_cell.i0ref_T_ref                                             = 353
        self.fuel_cell.current_density_limit_multiplier                        = 1
        self.fuel_cell.area_specific_mass                                      = 2.5 
        return 
        
    def energy_calc(self,state,bus,coolant_lines, t_idx, delta_t): 
        """Computes the state of the fuel cell battery cell. 
        """        
        if not (self.fuel_cell.type == "LT") or  (self.fuel_cell.type == "HT"): 
            raise ValueError('PEM type not supported, currently supported types are "LT" and "HT"')         
        
        stored_results_flag, stored_battery_tag = compute_fuel_cell_performance(self,state,bus,coolant_lines, t_idx,delta_t) 
        
        return stored_results_flag, stored_battery_tag
    
    def append_operating_conditions(self,segment,bus):  
        append_fuel_cell_conditions(self,segment,bus)  
        return
    
    def append_fuel_cell_segment_conditions(self,bus, conditions, segment):
        append_fuel_cell_segment_conditions(self,bus, conditions, segment)
        return 

    def reuse_stored_data(self,state,bus,stored_results_flag, stored_fuel_cell_tag):
        reuse_stored_fuel_cell_data(self,state,bus,stored_results_flag, stored_fuel_cell_tag)
        return       


# RCAIDE/Compoments/Propulsors/Converters/Ducted_Fan.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core              import Data
from .Converter                         import Converter
from RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.append_ducted_fan_conditions import  append_ducted_fan_conditions
import numpy as np
import scipy as sp
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Nacalle
# ----------------------------------------------------------------------------------------------------------------------  
class Ducted_Fan(Converter):
    """
    Ducted Fan Component Class

    This class models a ducted fan propulsion system with both rotor and stator components. 
    It inherits from the base Converter class and implements ducted-fan specific attributes and methods.

    Attributes
    ----------
    tag : str
        Identifier for the ducted fan component, defaults to 'ducted_fan'
    number_of_radial_stations : int
        Number of radial stations for blade element analysis, defaults to 20
    number_of_rotor_blades : int
        Number of rotor blades, defaults to 12
    tip_radius : float
        Outer radius of the rotor [m], defaults to 1.0
    hub_radius : float
        Inner radius of the rotor at hub [m], defaults to 0.1
    exit_radius : float
        Radius at the duct exit [m], defaults to 1.1
    blade_clearance : float
        Clearance between blade tip and duct wall [m], defaults to 0.01
    length : float
        Axial length of the ducted fan [m], defaults to 1.0
    fan_effectiveness : float
        Fan effectiveness factor [-], defaults to 1.1
    Cp_polynomial_coefficients : list
        Coefficients for power coefficient polynomial [-, -, -]
    Ct_polynomial_coefficients : list
        Coefficients for thrust coefficient polynomial [-, -, -]
    etap_polynomial_coefficients : list
        Coefficients for propulsive efficiency polynomial [-, -, -]
    fidelity : str
        Analysis fidelity level, either 'Blade_Element_Momentum_Theory' or 'Rankine_Froude_Momentum_Theory'
    orientation_euler_angles : list
        Default orientation angles of rotor [rad, rad, rad]
    rotor : Data
        Container for rotor geometry and performance data
    stator : Data
        Container for stator geometry and performance data
    cruise : Data
        Container for cruise design conditions

    Notes
    -----
    The ducted fan model includes detailed geometric parameters and performance characteristics
    for both the rotor and stator components. The model supports multiple fidelity levels
    and includes coordinate transformation capabilities for thrust vectoring analysis.

    **Major Assumptions**
        * Axisymmetric flow
        * Steady state operation
        * Incompressible flow for low-fidelity analysis
        * No radial flow
        * Uniform inflow
        * No blade-to-blade interaction

    """ 
    
    def __defaults__(self):
        """ This sets the default values for the component to function.
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        None
        """      
        
        self.tag                                   = 'ducted_fan'  
        self.number_of_radial_stations             = 20
        self.number_of_rotor_blades                = 12  
        self.tip_radius                            = 1.0
        self.hub_radius                            = 0.1
        self.exit_radius                           = 1.1
        self.blade_clearance                       = 0.01
        self.length                                = 1 
        self.fan_effectiveness                     = 1.1
        self.DFDC                                  = Data()
        self.DFDC.bin_name                         = 'dfdc'
        self.Cp_polynomial_coefficients            = [0.551,  0.0182, -0.0869]   
        self.Ct_polynomial_coefficients            = [0.4605,-0.0529, -0.1203]   
        self.etap_polynomial_coefficients          = [0.0653,4.1603 , -7.6128]  
        self.fidelity                              = 'Blade_Element_Momentum_Theory' # 'Rankine_Froude_Momentum_Theory'  
        self.orientation_euler_angles              = [0.,0.,0.]  # vector of angles defining default orientation of rotor
        self.rotor                                 = Data()
        self.stator                                = Data()
        self.rotor.percent_x_location              = 0.4
        self.stator.percent_x_location             = 0.7
        self.cruise                                = Data() 
        self.cruise.design_thrust                  = None
        self.cruise.design_power                   = None
        self.cruise.design_altitude                = None
        self.cruise.design_efficiency              = None  
        self.cruise.design_angular_velocity        = None 
        self.cruise.design_freestream_velocity     = None
        self.cruise.design_reference_velocity      = None 
        self.cruise.design_freestream_mach         = None  
        self.duct_airfoil                          = None
        self.hub_airfoil                           = None
      
    
    def append_duct_airfoil(self, airfoil):
        """
        Adds an airfoil to the ducted fan's duct section.

        Parameters
        ----------
        airfoil : Data
            Airfoil data container with aerodynamic properties for the duct section.
            Must be of type Data().

        Returns
        -------
        None

        Notes
        -----
        This method appends airfoil data to the duct_airfoil attribute of the ducted fan.
        The airfoil data is used to model the aerodynamic characteristics of the duct
        section, which influences the overall performance of the ducted fan system.

        Raises
        ------
        Exception
            If input airfoil is not of type Data()
        """

        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil') 

        # Store data
        self.duct_airfoil =  airfoil

        return
    

    def append_hub_airfoil(self, airfoil):
        """
        Adds an airfoil to the ducted fan's hub section.

        Parameters
        ----------
        airfoil : Data
            Airfoil data container with aerodynamic properties for the hub section.
            Must be of type Data().

        Returns
        -------
        None

        Notes
        -----
        This method appends airfoil data to the hub_airfoil attribute of the ducted fan.
        The airfoil data is used to model the aerodynamic characteristics of the hub
        section, which affects the flow field and performance of the ducted fan system.

        Raises
        ------
        Exception
            If input airfoil is not of type Data()
        """

        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil') 

        # Store data
        self.hub_airfoil =  airfoil

        return 

    def append_operating_conditions(ducted_fan,segment,energy_conditions,noise_conditions=None):  
        append_ducted_fan_conditions(ducted_fan,segment,energy_conditions,noise_conditions)
        return        
          
    def vec_to_vel(self):
        """
        Rotates from the ducted fan's vehicle frame to the ducted fan's velocity frame.

        Parameters
        ----------
        None

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from vehicle frame to velocity frame.

        Notes
        -----
        This method creates a rotation matrix for transforming coordinates between
        two reference frames of the ducted fan. When the ducted fan is axially
        aligned with the vehicle body:

        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The transformation is accomplished using a rotation of π radians about the Y-axis,
        represented as a rotation vector [0, π, 0].

        **Major Assumptions**
        * The ducted fan's default orientation is aligned with the vehicle body
        * Right-handed coordinate system is used
        """

        rot_mat = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        return rot_mat
    

    def body_to_prop_vel(self, commanded_thrust_vector):
        """
        Rotates from the system's body frame to the ducted fan's velocity frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from body frame to ducted fan velocity frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles.

        Notes
        -----
        This method performs a sequence of rotations to transform coordinates from
        the vehicle body frame to the ducted fan's velocity frame. The transformation
        sequence is:
        1. Body to vehicle frame (π rotation about Y-axis)
        2. Vehicle to ducted fan vehicle frame (includes thrust vector rotation)
        3. Ducted fan vehicle to ducted fan velocity frame

        Reference frames:
        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The complete transformation is computed as:
        rot_mat = (body_2_vehicle @ vehicle_2_duct_vec) @ duct_vec_2_duct_vel

        **Major Assumptions**
        * The ducted fan's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        * Matrix multiplication order preserves proper transformation sequence
        """

        # Go from velocity to vehicle frame
        body_2_vehicle = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()

        # Go from vehicle frame to ducted fan vehicle frame: rot 1 including the extra body rotation
        cpts       = len(np.atleast_1d(commanded_thrust_vector))
        rots       = np.array(self.orientation_euler_angles) * 1.
        rots       = np.repeat(rots[None,:], cpts, axis=0) 
        rots[:,1] += commanded_thrust_vector[:,0] 
        
        vehicle_2_duct_vec = sp.spatial.transform.Rotation.from_rotvec(rots).as_matrix()

        # GO from the ducted fan vehicle frame to the ducted fan velocity frame: rot 2
        duct_vec_2_duct_vel = self.vec_to_vel()

        # Do all the matrix multiplies
        rot1    = np.matmul(body_2_vehicle,vehicle_2_duct_vec)
        rot_mat = np.matmul(rot1,duct_vec_2_duct_vel)
 
        return rot_mat , rots


    def duct_vel_to_body(self, commanded_thrust_vector):
        """
        Rotates from the ducted fan's velocity frame to the system's body frame.

        Parameters
        ----------
        commanded_thrust_vector : ndarray
            Vector of commanded thrust angles [rad] for each time step.

        Returns
        -------
        rot_mat : ndarray
            3x3 rotation matrix transforming from ducted fan velocity frame to body frame.
        rots : ndarray
            Array of rotation vectors including commanded thrust angles.

        Notes
        -----
        This method performs the inverse transformation sequence of body_to_prop_vel.
        The transformation sequence is:
        1. Ducted fan velocity to ducted fan vehicle frame
        2. Ducted fan vehicle to vehicle frame (includes thrust vector rotation)
        3. Vehicle to body frame (π rotation about Y-axis)

        Reference frames:
        Velocity frame:
        * X-axis points out the nose
        * Z-axis points towards the ground
        * Y-axis points out the right wing

        Vehicle frame:
        * X-axis points towards the tail
        * Z-axis points towards the ceiling
        * Y-axis points out the right wing

        **Theory**
        The transformation is computed by inverting the rotation matrix from 
        body_to_prop_vel using:
        rot_mat = (body_2_duct_vel)^(-1)

        **Major Assumptions**
        * The ducted fan's default orientation is defined by orientation_euler_angles
        * Right-handed coordinate system is used
        * Thrust vector rotation is applied about the Y-axis
        """

        body2ductvel,rots = self.body_to_duct_vel(commanded_thrust_vector)

        r = sp.spatial.transform.Rotation.from_matrix(body2ductvel)
        r = r.inv()
        rot_mat = r.as_matrix()

        return rot_mat, rots
    
    def vec_to_duct_body(self,commanded_thrust_vector):
        rot_mat, rots =  self.duct_vel_to_body(commanded_thrust_vector) 
        return rot_mat, rots 


# RCAIDE/Library/Methods/Powertrain/Propulsors/Turbofan/design_turbofan.py 
# 
# Created:  Jul 2024, RCAIDE Team

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE 
from RCAIDE.Library.Methods.Powertrain.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Combustor          import compute_combustor_performance 
from RCAIDE.Library.Methods.Powertrain.Converters.Compressor         import compute_compressor_performance 
from RCAIDE.Library.Methods.Powertrain.Converters.Fan                import compute_fan_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan           import size_core 
from RCAIDE.Library.Methods.Powertrain                               import setup_operating_conditions 


# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Turbofan
# ---------------------------------------------------------------------------------------------------------------------- 
def design_turbofan(turbofan):
    """
    Computes performance properties of a turbofan engine at the design point by linking
    and analyzing the thermodynamic cycle of its components.
    
    Parameters
    ----------
    turbofan : RCAIDE.Library.Components.Propulsors.Turbofan
        Turbofan engine component with the following attributes:
            - tag : str
                Identifier for the turbofan
            - design_mach_number : float
                Design Mach number
            - design_altitude : float
                Design altitude [m]
            - design_isa_deviation : float
                ISA temperature deviation at design point [K]
            - working_fluid : Data
                Working fluid properties object
            - reference_temperature : float
                Reference temperature for mass flow scaling [K]
            - reference_pressure : float
                Reference pressure for mass flow scaling [Pa]
            - bypass_ratio : float
                Bypass ratio of the turbofan
            - ram : Data
                Ram component
                    - tag : str
                        Identifier for the ram
            - inlet_nozzle : Data
                Inlet nozzle component
                    - tag : str
                        Identifier for the inlet nozzle
            - fan : Data
                Fan component
                    - tag : str
                        Identifier for the fan
            - low_pressure_compressor : Data
                Low pressure compressor component
                    - tag : str
                        Identifier for the low pressure compressor
            - high_pressure_compressor : Data
                High pressure compressor component
                    - tag : str
                        Identifier for the high pressure compressor
            - combustor : Data
                Combustor component
                    - tag : str
                        Identifier for the combustor
            - high_pressure_turbine : Data
                High pressure turbine component
                    - tag : str
                        Identifier for the high pressure turbine
            - low_pressure_turbine : Data
                Low pressure turbine component
                    - tag : str
                        Identifier for the low pressure turbine
            - core_nozzle : Data
                Core nozzle component
                    - tag : str
                        Identifier for the core nozzle
            - fan_nozzle : Data
                Fan nozzle component
                    - tag : str
                        Identifier for the fan nozzle
    
    Returns
    -------
    None
    
    Notes
    -----
    This function performs a complete design analysis of a turbofan engine by:
        1. Setting up atmospheric conditions at the design point
        2. Creating a mission segment for the design point
        3. Sequentially analyzing each component in the engine's thermodynamic cycle
        4. Linking the output conditions of each component to the input conditions of the next
        5. Sizing the core flow to meet the design requirements
        6. Computing sea level static performance
    
    The function follows this sequence for component analysis:
        1. Ram (inlet)
        2. Inlet nozzle
        3. Fan
        4. Low pressure compressor
        5. High pressure compressor
        6. Combustor
        7. High pressure turbine
        8. Low pressure turbine
        9. Core nozzle
        10. Fan nozzle
    
    **Major Assumptions**
        * US Standard Atmosphere 1976
        * Steady state operation
        * One-dimensional flow through components
        * Adiabatic components except for the combustor
        * Perfect gas behavior with variable properties
    
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", 2nd Edition, AIAA Education Series, 2005. https://soaneemrana.org/onewebmedia/ELEMENTS%20OF%20GAS%20TURBINE%20PROPULTION2.pdf
    [2] Cantwell, B., "AA283 Course Notes", Stanford University. https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_BOOK/AA283_Aircraft_and_Rocket_Propulsion_BOOK_Brian_J_Cantwell_May_28_2024.pdf
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.compute_turbofan_performance
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.size_core
    """
    # check if mach number and temperature are passed
    if(turbofan.design_mach_number==None) and (turbofan.design_altitude==None): 
        raise NameError('The sizing conditions require an altitude and a Mach number') 
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turbofan.design_altitude,turbofan.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity           
        U   = a*turbofan.design_mach_number
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turbofan.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turbofan.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turbofan.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turbofan.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turbofan.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turbofan.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(U) 
     
    segment                  = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions = conditions 
    turbofan.append_operating_conditions(segment,conditions.energy,conditions.noise)
                    
    ram                       = turbofan.ram
    inlet_nozzle              = turbofan.inlet_nozzle
    fan                       = turbofan.fan
    low_pressure_compressor   = turbofan.low_pressure_compressor
    high_pressure_compressor  = turbofan.high_pressure_compressor
    combustor                 = turbofan.combustor
    high_pressure_turbine     = turbofan.high_pressure_turbine
    low_pressure_turbine      = turbofan.low_pressure_turbine
    core_nozzle               = turbofan.core_nozzle
    fan_nozzle                = turbofan.fan_nozzle 
    bypass_ratio              = turbofan.bypass_ratio  

    # unpack component conditions
    turbofan_conditions     = conditions.energy.propulsors[turbofan.tag]
    ram_conditions          = conditions.energy.converters[ram.tag]    
    inlet_nozzle_conditions = conditions.energy.converters[inlet_nozzle.tag]
    fan_conditions          = conditions.energy.converters[fan.tag]    
    lpc_conditions          = conditions.energy.converters[low_pressure_compressor.tag]
    hpc_conditions          = conditions.energy.converters[high_pressure_compressor.tag]
    combustor_conditions    = conditions.energy.converters[combustor.tag] 
    lpt_conditions          = conditions.energy.converters[low_pressure_turbine.tag]
    hpt_conditions          = conditions.energy.converters[high_pressure_turbine.tag]
    core_nozzle_conditions  = conditions.energy.converters[core_nozzle.tag]
    fan_nozzle_conditions   = conditions.energy.converters[fan_nozzle.tag]    
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turbofan.working_fluid
    
    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,conditions)
    
    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid
    
    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,conditions)
    
    # Step 5: Link the fan to the inlet nozzle
    fan_conditions.inputs.stagnation_temperature                      = inlet_nozzle_conditions.outputs.stagnation_temperature
    fan_conditions.inputs.stagnation_pressure                         = inlet_nozzle_conditions.outputs.stagnation_pressure
    fan_conditions.inputs.static_temperature                          = inlet_nozzle_conditions.outputs.static_temperature
    fan_conditions.inputs.static_pressure                             = inlet_nozzle_conditions.outputs.static_pressure
    fan_conditions.inputs.mach_number                                 = inlet_nozzle_conditions.outputs.mach_number  
    fan.working_fluid                                                 = inlet_nozzle.working_fluid
     
    # Step 6: Compute flow through the fan
    compute_fan_performance(fan,conditions)    

    # Step 7: Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature                      = fan_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure                         = fan_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature                          = fan_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure                             = fan_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number                                 = fan_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid                             = fan.working_fluid
    low_pressure_compressor.reference_temperature                     = turbofan.reference_temperature
    low_pressure_compressor.reference_pressure                        = turbofan.reference_pressure
    
    # Step 8: Compute flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,conditions)
    
    # Step 9: Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature                      = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure                         = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature                          = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure                             = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number                                 = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid                            = low_pressure_compressor.working_fluid  
    high_pressure_compressor.reference_temperature                    = turbofan.reference_temperature
    high_pressure_compressor.reference_pressure                       = turbofan.reference_pressure  
    
    # Step 10: Compute flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,conditions)

    # Step 11: Link the combustor to the high pressure compressor    
    combustor_conditions.inputs.stagnation_temperature                = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                                           = high_pressure_compressor.working_fluid     
    
    # Step 12: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,conditions)
    
    # Step 13: Link the high pressure turbione to the combustor
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number       
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs  
    hpt_conditions.inputs.fan                       = fan_conditions.outputs 
    hpt_conditions.inputs.bypass_ratio              = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid    
    
    # Step 14: Compute flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,conditions)
            
    # Step 15: Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure  
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number    
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid     
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs 
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.fan                        = fan_conditions.outputs 
    lpt_conditions.inputs.bypass_ratio               = bypass_ratio 
    
    # Step 16: Compute flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,conditions)
    
    # Step 17: Link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid 
    
    # Step 18: Compute flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,conditions)
   
    # Step 19: Link the fan nozzle to the fan
    fan_nozzle_conditions.inputs.stagnation_temperature     = fan_conditions.outputs.stagnation_temperature
    fan_nozzle_conditions.inputs.stagnation_pressure        = fan_conditions.outputs.stagnation_pressure
    fan_nozzle_conditions.inputs.static_temperature         = fan_conditions.outputs.static_temperature
    fan_nozzle_conditions.inputs.static_pressure            = fan_conditions.outputs.static_pressure  
    fan_nozzle_conditions.inputs.mach_number                = fan_conditions.outputs.mach_number   
    fan_nozzle.working_fluid                                = fan.working_fluid
    
    # Step 20: Compute flow through the fan nozzle
    compute_expansion_nozzle_performance(fan_nozzle,conditions)
     
    # Step 21: Link the turbofan to outputs from various compoments    
    turbofan_conditions.bypass_ratio                             = bypass_ratio
    turbofan_conditions.fan_nozzle_exit_velocity                 = fan_nozzle_conditions.outputs.velocity
    turbofan_conditions.fan_nozzle_area_ratio                    = fan_nozzle_conditions.outputs.area_ratio  
    turbofan_conditions.fan_nozzle_static_pressure               = fan_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_area_ratio                   = core_nozzle_conditions.outputs.area_ratio 
    turbofan_conditions.core_nozzle_static_pressure              = core_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_exit_velocity                = core_nozzle_conditions.outputs.velocity 
    turbofan_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 
    turbofan_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbofan_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure
    turbofan_conditions.flow_through_core                        = 1./(1.+bypass_ratio) #scaled constant to turn on core thrust computation
    turbofan_conditions.flow_through_fan                         = bypass_ratio/(1.+bypass_ratio) #scaled constant to turn on fan thrust computation        

    # Step 22: Size the core of the turbofan  
    size_core(turbofan,conditions) 
    
    # Step 23: Static Sea Level Thrust  
    atmo_data_sea_level   = atmosphere.compute_values(0.0,0.0)   
    V                     = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state       = setup_operating_conditions(turbofan, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.propulsors[turbofan.tag].throttle[:,0] = 1.0  
    sls_T,_,sls_P,_,_,_                          = turbofan.compute_performance(operating_state) 
    turbofan.sealevel_static_thrust              = sls_T[0][0]
    turbofan.sealevel_static_power               = sls_P[0][0]
     
    return 


# RCAIDE/Methods/Powertrain/Sources/Fuel_Cells/Larminie_Model/compute_voltage.py
#  
# Created: Jan 2025, M. Clarke
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Find Voltage Larminie
# ----------------------------------------------------------------------------------------------------------------------
def compute_voltage(fuel_cell, current_density):
    """
    Calculates fuel cell voltage based on current density using the Larminie-Dicks model.
    
    Parameters
    ----------
    fuel_cell : RCAIDE.Components.Energy.Converters.Fuel_Cell
        The fuel cell component containing electrochemical parameters
            - r : float
                Area-specific resistance [Ohms*cm²]
            - A1 : float
                Tafel slope [V]
            - m : float
                Mass transport loss coefficient [V]
            - n : float
                Mass transport loss exponential coefficient [cm²/A]
            - Eoc : float
                Open circuit voltage [V]
    current_density : float or array
        Current density [A/m²]
        
    Returns
    -------
    v : float or array
        Cell voltage [V]
    
    Notes
    -----
    This function implements the Larminie-Dicks semi-empirical model to calculate
    fuel cell voltage as a function of current density. The model accounts for
    activation losses, ohmic losses, and concentration losses.
    
    **Major Assumptions**
        * Voltage curve follows the Larminie-Dicks model form
        * Steady-state operation (no transient effects)
        * Uniform current distribution across the cell
        * Constant temperature operation
    
    **Theory**
    
    The Larminie-Dicks model calculates cell voltage as:
    
    .. math::
        V = E_{oc} - r \\cdot i - A_1 \\ln(i) - m \\exp(n \\cdot i)
    
    where:
        - :math:`E_{oc}` is the open circuit voltage
        - r is the area-specific resistance
        - A_1 is the Tafel slope for activation losses
        - m and n are parameters for mass transport losses
        - i is the current density
    
    References
    ----------
    [1] Larminie, J., & Dicks, A. (2003). Fuel Cell Systems Explained (2nd ed.). John Wiley & Sons Ltd.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Larminie_Model.compute_power
    """
    r   = fuel_cell.r/(1000*(Units.cm**2))
    Eoc = fuel_cell.Eoc 
    A1  = fuel_cell.A1  
    m   = fuel_cell.m   
    n   = fuel_cell.n   
    
    i1 = current_density/(0.001/(Units.cm**2.)) # current density(mA cm^-2)
    v  = Eoc-r*i1-A1*np.log(i1)-m*np.exp(n*i1)     #useful voltage vector

    return v


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/__init__.py
# 

"""
Methods for designing rotors in aircraft propulsion systems.

This module provides functions for setting up and executing the design process for
various types of rotors, including propellers, lift rotors, and prop rotors. It includes
methods for blade geometry setup, optimization parameter configuration, design procedure
definition, and post-optimization parameter setting.

The design process typically involves:
    1. Setting up the blade geometry with appropriate airfoil sections and twist distribution
    2. Configuring the optimization parameters and constraints
    3. Defining the design procedure and analysis methods
    4. Running the optimization to find the optimal blade design
    5. Setting the optimized parameters in the rotor object

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .blade_geometry_setup         import blade_geometry_setup
from .optimization_setup           import optimization_setup
from .procedure_setup              import procedure_setup
from .set_optimized_parameters     import set_optimized_parameters




# RCAIDE/Library/Methods/Powertrain/Propulsors/Electric_Rotor_Propulsor/unpack_electric_rotor_unknowns.py

# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  unpack electric rotor network unknowns 
# ----------------------------------------------------------------------------------------------------------------------  

def unpack_electric_rotor_unknowns(propulsor,segment): 
    '''
    Unpack residuals for electric rotor and assigns them to the specfic
    compoment each interation of the mission solver   
    '''
    motor   = propulsor.motor   
    motor_conditions = segment.state.conditions.energy.converters[motor.tag]
    motor_conditions.inputs.current = segment.state.unknowns[propulsor.tag + '_motor_current'] 
    return 


# RCAIDE/Library/Missions/Segments/Cruise/Constant_Dynamic_Pressure_Constant_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
import RCAIDE 

# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """
    Initializes conditions for constant dynamic pressure cruise at fixed altitude

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

        - altitude : float
            Cruise altitude [m]. If not specified, the altitude from the previous segment is used.
        - distance : float
            Ground distance to cover [m]
        - dynamic_pressure : float
            Dynamic pressure to maintain [Pa]
        - sideslip_angle : float
            Aircraft sideslip angle [rad]
        - state:
            numerics.dimensionless.control_points : array
                Discretization points [-]
            conditions : Data
                State conditions container
        - analyses:
            atmosphere : Model
                Atmospheric model for property calculations

    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.position_vector [m]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.time [s]
    
    Notes
    -----
    This function sets up the initial conditions for a cruise segment with constant
    dynamic pressure and constant altitude. The airspeed is determined from the
    dynamic pressure constraint.

    **Calculation Process**
        1. Get atmospheric properties at altitude
        2. Calculate true airspeed from dynamic pressure:
            V = sqrt(2q/ρ) where:
                - q is dynamic pressure
                - ρ is air density
        3. Calculate time required based on distance and speed
        4. Discretize time points
        5. Decompose velocity into components using sideslip angle

    **Major Assumptions**
        * Constant dynamic pressure
        * Constant altitude
        * Standard atmosphere model
        * Small angle approximations
        * Quasi-steady flight
        * No wind effects

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """        
    
    # unpack
    alt        = segment.altitude
    xf         = segment.distance
    q          = segment.dynamic_pressure
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions   
    
    # Update freestream to get density
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment)
    rho        = conditions.freestream.density[:,0]   
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]      
    
    # check for initial velocity
    if q is None: 
        if not segment.state.initials: raise AttributeError('dynamic pressure not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
    else: 
        # compute speed, constant with constant altitude
        air_speed = np.sqrt(q/(rho*0.5))
    
    # dimensionalize time
    t_initial = conditions.frames.inertial.time[0,0]
    t_final   = xf / air_speed + t_initial
    t_nondim  = segment.state.numerics.dimensionless.control_points
    time      = t_nondim * (t_final-t_initial) + t_initial
    v_x       = np.cos(beta)*air_speed 
    v_y       = np.sin(beta)*air_speed 
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import numpy as np
from RCAIDE.Library.Methods.Geometry.LOPA.compute_layout_of_passenger_accommodations import  compute_layout_of_passenger_accommodations

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------

def fuselage_planform(fuselage, circular_cross_section = True):
    """Calculates fuselage geometry values

    Assumptions:
    None

    Source:
    http://adg.stanford.edu/aa241/drag/wettedarea.html

    Inputs:
    fuselage.
      num_coach_seats       [-] 
      fineness.nose         [-]
      fineness.tail         [-] 
      width                 [m]
      heights.maximum       [m]

    Outputs:
    fuselage.
      lengths.nose          [m]
      lengths.tail          [m]
      lengths.cabin         [m]
      lengths.total         [m]
      areas.wetted          [m]
      areas.front_projected [m]
      effective_diameter    [m]

    Properties Used:
    N/A
    """
    # size cabins 
    compute_layout_of_passenger_accommodations(fuselage)
     
    nose_fineness   = fuselage.fineness.nose
    tail_fineness   = fuselage.fineness.tail
    
    cabin_length    = 0
    fuselage_width  = 0
    for cabin in fuselage.cabins: 
        cabin_length += cabin.length
        fuselage_width =  np.maximum(fuselage_width, cabin.width)
        
    nose_length     = nose_fineness * fuselage_width
    tail_length     = tail_fineness * fuselage_width 
    fuselage_length = cabin_length + nose_length + tail_length 
    fuselage.lengths.total = fuselage_length    

    if circular_cross_section:
        fuselage_height = fuselage_width
    else: 
        fuselage_height = fuselage.heights.maximum
            
    wetted_area = 0.0 
    # model constant fuselage cross section as an ellipse
    # approximate circumference http://en.wikipedia.org/wiki/Ellipse#Circumference
    a = fuselage_width/2.
    b = fuselage_height/2.
    A = np.pi * a * b  # area
    R = (a-b)/(a+b) # effective radius
    C = np.pi*(a+b)*(1.+ ( 3*R**2 )/( 10+np.sqrt(4.-3.*R**2) )) # circumference
    
    wetted_area += C * cabin_length
    cross_section_area = A
    
    # approximate nose and tail wetted area
    # http://adg.stanford.edu/aa241/drag/wettedarea.html
    Deff = (a+b)*(64.-3.*R**4)/(64.-16.*R**2)
    wetted_area += 0.75*np.pi*Deff * (nose_length + tail_length)
    
    # update
    fuselage.lengths.nose          = nose_length
    fuselage.lengths.tail          = tail_length
    fuselage.lengths.cabin         = cabin_length
    fuselage.lengths.total         = fuselage_length
    fuselage.areas.wetted          = wetted_area
    fuselage.areas.front_projected = cross_section_area
    fuselage.effective_diameter    = Deff 
    return fuselage



# RCAIDE/Library/Components/Propulsors/Turboelectric_Generator.py
# 
#  
# Created:  Jan 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
## RCAIDE imports
from RCAIDE.Framework.Core                  import Data 
from .Converter                             import Converter
from RCAIDE.Library.Components.Powertrain.Converters.Turboshaft    import Turboshaft 
from RCAIDE.Library.Components.Powertrain.Converters.DC_Generator  import DC_Generator 
from RCAIDE.Library.Methods.Powertrain.Converters.Turboelectric_Generator.append_turboelectric_generator_conditions      import append_turboelectric_generator_conditions  
from RCAIDE.Library.Methods.Powertrain.Converters.Turboelectric_Generator.compute_turboelectric_generator_performance    import compute_turboelectric_generator_performance, reuse_stored_turboelectric_generator_data
 
# ----------------------------------------------------------------------
#  Turboelectric_Generator
# ----------------------------------------------------------------------
class Turboelectric_Generator(Converter):
    """
    A Turboelectric_Generator propulsion system model that simulates the performance of a Turboelectric_Generator engine.
   

    Attributes
    ----------
    tag : str
        Identifier for the shaft engine. Default is 'turboshaft'. 
        
    turboshaft : Component
        Turboshaft component. Default is the Turboshaft Class.
        
    generator : Component
        Generator component. Default is DC_Generator Class.
        
    gearbox : Component
        Gearbox data structure. Default is None. 
        
    inverse_calculation : Component
        Flag that determines the how calculations are performed. Default is False    

    Notes
    -----
    The Turboelectric_Generator class inherits from the Turboshaft class and implements
    methods for computing Turboelectric_Generator engine performance. Unlike other gas turbine
    engines that produce thrust, a Turboelectric_Generator engine's primary output is shaft
    power, typically used to drive a helicopter rotor or other mechanical systems. 

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Propulsors.Turboshaft 
    """ 
    def __defaults__(self):
        # setting the default values
        self.tag                       = 'turboelectric_generator'
        self.turboshaft                = Turboshaft()
        self.generator                 = DC_Generator()
        self.gearbox                   = Data()
        self.gearbox.gear_ratio        = None  
        self.inverse_calculation       = False

    def append_operating_conditions(self,segment,energy_conditions,noise_conditions=None): 
        """
        Appends operating conditions to the segment.
        """  
        append_turboelectric_generator_conditions(self,segment,energy_conditions) 
        return
 
    def compute_performance(self,state,fuel_line = None,bus = None):
        """
        Computes Turboelectric_Generator performance including power.
        """
        P_mech,P_elec,stored_results_flag,stored_propulsor_tag =  compute_turboelectric_generator_performance(self,state,fuel_line, bus)
        return P_mech,P_elec,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(turboelectric_generator,state,stored_propulsor_tag):
        power  = reuse_stored_turboelectric_generator_data(turboelectric_generator,state,stored_propulsor_tag)
        return power 



# RCAIDE/Framework/Analyses/Weights/Conventional.py
#
# Created:  Feb 2025, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

import RCAIDE
from RCAIDE.Framework.Core import Data 
from .Weights import Weights

# ----------------------------------------------------------------------------------------------------------------------
#  Transport Weights Analysis
# ----------------------------------------------------------------------------------------------------------------------
class Conventional(Weights):
    """ This is class that evaluates the weight of ******* class aircraft

    Assumptions:
        None

    Source:
        N/A

    Inputs:
        None

    Outputs:
        None

    Properties Used:
         N/A
    """

    def __defaults__(self):
        """This sets the default values and methods for the tube and wing
        aircraft weight analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """
        self.method                                        = 'FLOPS'
        self.aircraft_type                                 = 'Transport'
        self.propulsion_architecture                       = 'Conventional'
                 
        self.settings.use_max_fuel_weight                  = True 
        self.settings.advanced_composites                  = False
        self.settings.fuselage_mounted_landing_gear_factor = 1.12 # assumes fuselage mounted landing gear. Change to 1 if False
        self.settings.cargo_doors_number                   = 1 # 0 if no cargo doors, 1 if 1 cargo door, 2 if 2 cargo doors
        self.settings.cargo_doors_clamshell                = False # True if clamshell cargo doors, False if not
        
        self.settings.weight_reduction_factors             = Data()  
        self.settings.weight_reduction_factors.main_wing   = 0.  # Reduction factors are proportional (.1 is a 10% weight reduction)
        self.settings.weight_reduction_factors.fuselage    = 0.  # Reduction factors are proportional (.1 is a 10% weight reduction)
        self.settings.weight_reduction_factors.empennage   = 0.  # applied to horizontal and vertical stabilizers
        
        # FLOPS settings
        self.settings.FLOPS                                = Data() 
        self.settings.FLOPS.complexity                     = 'Simple' 
        self.settings.FLOPS.aeroelastic_tailoring_factor   = 0.   # Aeroelastic tailoring factor [0 no aeroelastic tailoring, 1 maximum aeroelastic tailoring] 
        self.settings.FLOPS.strut_braced_wing_factor       = 0.   # Wing strut bracing factor [0 for no struts, 1 for struts]
        
     


# RCAIDE/Library/Methods/Energy/Converters/Turboshaft/design_turboshaft.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports     
import RCAIDE
from RCAIDE.Framework.Core                                           import Data
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Powertrain.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft         import size_core  
from RCAIDE.Library.Methods.Powertrain                               import setup_operating_conditions 

# Python package imports   
import numpy                                                                as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Turboshaft
# ----------------------------------------------------------------------------------------------------------------------
def design_turboshaft(turboshaft):  
    """
    Designs and sizes a turboshaft engine based on design point conditions and performance requirements.

    Parameters
    ----------
    turboshaft : Turboshaft
        Turboshaft engine object containing design parameters and components
            - design_mach_number : float
                Design point Mach number
            - design_altitude : float
                Design point altitude [m]
            - design_isa_deviation : float
                ISA temperature deviation [K]
            - working_fluid : FluidProperties
                Working fluid properties object
            - Components:
                - ram : Ram
                - inlet_nozzle : Compression_Nozzle
                - compressor : Compressor
                - combustor : Combustor
                - high_pressure_turbine : Turbine
                - low_pressure_turbine : Turbine
                - core_nozzle : Expansion_Nozzle

    Returns
    -------
    None
        Results are stored in turboshaft attributes:
            - design_thrust_specific_fuel_consumption : float
                TSFC at design point [kg/N/s]
            - design_non_dimensional_thrust : float
                Non-dimensional thrust [-]
            - design_core_mass_flow_rate : float
                Core mass flow rate [kg/s]
            - design_fuel_flow_rate : float
                Fuel flow rate [kg/s]
            - design_power : float
                Power output [W]
            - design_specific_power : float
                Specific power [W/kg]
            - design_power_specific_fuel_consumption : float
                Power specific fuel consumption [kg/W/s]
            - design_thermal_efficiency : float
                Thermal efficiency [-]
            - design_propulsive_efficiency : float
                Propulsive efficiency [-]

    Notes
    -----
    The function performs the following steps:
        1. Computes atmospheric conditions at design altitude
        2. Initializes design point conditions
        3. Analyzes flow through each component sequentially:
            - Ram inlet
            - Inlet nozzle
            - Compressor
            - Combustor
            - High pressure turbine
            - Low pressure turbine
            - Core nozzle
        4. Sizes core components
        5. Computes sea level static performance

    **Major Assumptions**
        * Standard atmospheric conditions (with possible ISA deviation)
        * Steady state operation
        * Perfect gas behavior
        * Adiabatic component processes except combustor
        * No bleed air extraction
        * Design point sizing determines component characteristics

    **Theory**

    The design process follows standard gas turbine cycle analysis principles.
    Component sizing is based on achieving required power output while maintaining
    appropriate component matching throughout the engine.

    **Extra modules required**
        * RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976
        * RCAIDE.Library.Attributes.Planets.Earth

    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Ram.compute_ram_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Combustor.compute_combustor_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Compressor.compute_compressor_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Turbine.compute_turbine_performance
    """
    
    #check if mach number and temperature are passed
    if(turboshaft.design_mach_number==None or turboshaft.design_altitude==None):
        
        #raise an error
        raise NameError('The sizing conditions require an altitude and a Mach number')
    
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turboshaft.design_altitude,turboshaft.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity   
    
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turboshaft.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turboshaft.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turboshaft.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turboshaft.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turboshaft.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turboshaft.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(a*turboshaft.design_mach_number)
         
          
    fuel_line                = RCAIDE.Library.Components.Powertrain.Distributors.Fuel_Line()    # may not need
    
    segment                  = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions = conditions
    turboshaft.append_operating_conditions(segment,conditions.energy,conditions.noise)  
            
    ram                     = turboshaft.ram
    inlet_nozzle            = turboshaft.inlet_nozzle
    compressor              = turboshaft.compressor
    combustor               = turboshaft.combustor
    high_pressure_turbine   = turboshaft.high_pressure_turbine
    low_pressure_turbine    = turboshaft.low_pressure_turbine
    core_nozzle             = turboshaft.core_nozzle
    
    turboshaft_conditions   = conditions.energy.converters[turboshaft.tag]
    ram_conditions          = conditions.energy.converters[ram.tag]     
    inlet_nozzle_conditions = conditions.energy.converters[inlet_nozzle.tag]
    core_nozzle_conditions  = conditions.energy.converters[core_nozzle.tag] 
    compressor_conditions   = conditions.energy.converters[compressor.tag]  
    combustor_conditions    = conditions.energy.converters[combustor.tag]
    lpt_conditions          = conditions.energy.converters[low_pressure_turbine.tag]
    hpt_conditions          = conditions.energy.converters[high_pressure_turbine.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turboshaft.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                             = inlet_nozzle.working_fluid 
    compressor.reference_temperature                     = turboshaft.reference_temperature
    compressor.reference_pressure                        = turboshaft.reference_pressure

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,conditions)
    
    # Step 11: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                                           = compressor.working_fluid 
    compressor.reference_temperature                                  = turboshaft.reference_temperature
    compressor.reference_pressure                                     = turboshaft.reference_pressure  
    
    # Step 12: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,conditions)

    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                = compressor_conditions.outputs 
    hpt_conditions.inputs.bypass_ratio              = 0.0
    hpt_conditions.inputs.fan                       = Data()
    hpt_conditions.inputs.fan.work_done             = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid
    
    compute_turbine_performance(high_pressure_turbine,conditions)
            
    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature              = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure                 = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature                  = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure                     = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                         = hpt_conditions.outputs.mach_number  
    low_pressure_turbine.working_fluid                        = high_pressure_turbine.working_fluid    
    lpt_conditions.inputs.compressor                          = Data()
    lpt_conditions.inputs.compressor.work_done                = 0.0 
    lpt_conditions.inputs.compressor.external_shaft_work_done = 0.0 
    lpt_conditions.inputs.fuel_to_air_ratio                   = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio                        = 0.0
    lpt_conditions.inputs.fan                                 = Data()
    lpt_conditions.inputs.fan.work_done                       = 0.0
     
    compute_turbine_performance(low_pressure_turbine,conditions)
    
    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid  
    
    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,conditions)

    # compute the thrust using the thrust component
    #link the thrust component to the core nozzle
    turboshaft_conditions.core_exit_velocity                  = core_nozzle_conditions.outputs.velocity
    turboshaft_conditions.core_area_ratio                     = core_nozzle_conditions.outputs.area_ratio
    turboshaft_conditions.core_nozzle                         = core_nozzle_conditions.outputs
    
    #link the thrust component to the combustor
    turboshaft_conditions.fuel_to_air_ratio                   = combustor_conditions.outputs.fuel_to_air_ratio
    
    #link the thrust component to the low pressure compressor
    turboshaft_conditions.combustor_stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    turboshaft_conditions.stag_temp_lpt_exit                  = compressor_conditions.inputs.stagnation_temperature
    turboshaft_conditions.stag_press_lpt_exit                 = compressor_conditions.inputs.stagnation_pressure 
    turboshaft_conditions.total_temperature_reference         = compressor_conditions.inputs.stagnation_temperature
    turboshaft_conditions.total_pressure_reference            = compressor_conditions.inputs.stagnation_pressure  

    #compute the power
    turboshaft_conditions.fan_nozzle                          = Data()
    turboshaft_conditions.fan_nozzle.velocity                 = 0.0
    turboshaft_conditions.fan_nozzle.area_ratio               = 0.0
    turboshaft_conditions.fan_nozzle.static_pressure          = 0.0
    turboshaft_conditions.bypass_ratio                        = 0.0
    turboshaft_conditions.flow_through_core                   = 1.0 #scaled constant to turn on core power computation
    turboshaft_conditions.flow_through_fan                    = 0.0 #scaled constant to turn on fan power computation      
    
    # Step 25: Size the core of the turboshaft  
    size_core(turboshaft,conditions)
    
    # Step 26: Static Sea Level Thrust  
    atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data_sea_level  = atmosphere.compute_values(0.0,0.0)   
    V                    = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state      = setup_operating_conditions(turboshaft, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.converters[turboshaft.tag].throttle[:,0] = 1.0  
    sls_P,_,_                                                       = turboshaft.compute_performance(operating_state,fuel_line) 
    turboshaft.sealevel_static_power                                = sls_P[0][0]
     
    return      
  


# RCAIDE/Library/Methods/Mass_Properties/Weight_Buildups/Electric/VTOL/Physics_Based/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_boom_weight                   import compute_boom_weight
from .compute_fuselage_weight               import compute_fuselage_weight
from .compute_operating_empty_weight        import compute_operating_empty_weight
from .compute_wing_weight                   import compute_wing_weight
from .dynamo_supply_mass_estimation         import dynamo_supply_mass_estimation



# RCAIDE/Library/Methods/Powertrain/Propulsors/Electric_Rotor_Propulsor/unpack_electric_ducted_fan_unknowns.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  unpack electric ducted_fan network unknowns 
# ----------------------------------------------------------------------------------------------------------------------  

def unpack_electric_ducted_fan_unknowns(propulsor,segment):
    """Unpack electric ducted fan unknowns and assigns them to the specfic
    compoment each interation of the mission solver
    """
    motor      = propulsor.motor  
    motor_conditions = segment.state.conditions.energy.converters[motor.tag]
    motor_conditions.inputs.current = segment.state.unknowns[propulsor.tag + '_motor_current'] 
    return 